import numpy as np
from spectral.spherical.swsh import Yslm_vec
from waveformtools.dataIO import construct_mode_list
from waveformtools.waveformtools import message
from waveformtools.diagnostics import method_info

# from numba import int32, float64, complex128, uint16, types, typed,
# from numba.types import List, Dict


# key and value types
# emode_list = [[2, 2], [0, 0]]
# emode_dict = {''}

# spec =[
#    ('modes_data', complex128[:]),
#    ('zero_modes', List),
#    ('non_zero_modes', List),
#    ('spin_weight', int32),
#    ('ell_max', uint16),
#    ('modes_list', list),
#    ('modes_dict', )
# ]


class SingleMode:
    """Storage and handling of a single mode"""

    def __init__(
        self,
        modes_data=None,
        zero_modes=None,
        non_zero_modes=None,
        spin_weight=0,
        ell_max=8,
        modes_list=None,
        modes_dict=None,
        tol=1e-8,
        extra_mode_axes_shape=None,
        Grid=None,
        vec_modes=None,
        func=None,
        label=None,
        sine_power=0,
    ):
        self._label = label
        self._modes_data = modes_data
        self._zero_modes = zero_modes
        self._non_zero_modes = non_zero_modes
        self._spin_weight = spin_weight
        self._ell_max = ell_max
        self._modes_list = modes_list
        self._tol = tol
        self._extra_mode_axes_shape = extra_mode_axes_shape
        self._Grid = Grid
        self._vec_modes = vec_modes
        self._func = func
        self._sine_power = sine_power

        if (np.array(modes_data) == np.array(None)).all():
            created = False
        else:
            created = True

        if isinstance(modes_dict, dict):
            message(
                "Creating SingleMode obj from modes dict...",
                message_verbosity=2,
            )

            ell_keys_str = list(modes_dict.keys())

            ell_keys_dict = [int(item[1:]) for item in ell_keys_str]

            ell_max_dict = max(ell_keys_dict)

            message(
                "Parsed dict ell_max as", ell_max_dict, message_verbosity=2
            )

            # if self._ell_max < ell_max:
            self._ell_max = ell_max_dict

            if not self._modes_list:
                self._modes_list = construct_mode_list(
                    ell_max=self.ell_max, spin_weight=self.spin_weight
                )

            self.create_modes_array()
            created = True
            for ell, emm_list in self.modes_list:
                for emm in emm_list:
                    value = modes_dict[f"l{ell}"][f"m{emm}"]
                    self.set_mode_data(ell=ell, emm=emm, value=value)

        if not created:
            self.create_modes_array()

        if Grid is None:
            from spectral.spherical.grids import GLGrid

            self._Grid = GLGrid(L=ell_max)

        self._modes_spherepack = None

        if not (np.array(self.vec_modes) == np.array(None)).all():
            self.construct_from_vec_modes()

        self.St, _ = self.Grid.meshgrid

    @property
    def label(self):
        return self._label

    @property
    def vec_modes(self):
        return self._vec_modes

    @property
    def extra_mode_axes_shape(self):
        return self._extra_mode_axes_shape

    @property
    def tol(self):
        return self._tol

    @property
    def Grid(self):
        return self._Grid

    @property
    def func(self):
        return self._func

    @property
    def sine_power(self):
        """This is just kept for
        bookkeeping purposes."""
        return self._sine_power

    @property
    def modes_data(self):
        return self._modes_data

    @property
    def modes_spherepack(self):
        """Get the modes in spherepack convention. Note that
        this assumes expansion of a real function in real spherical
        harmonics"""
        if (np.array(self._modes_spherepack) == np.array(None)).any():

            self.get_in_spherepack_convention()

        return self._modes_spherepack

    def get_modes_dict(self, modes_list=None):
        """Get a dictionary representation of the
        modes requested"""

        modes_dict = {}

        if not modes_list:
            message("Create SingleMode modes list", message_verbosity=2)
            modes_list = self.modes_list
            message(f"mode list {modes_list}", message_verbosity=3)

        for ell, emm_list in modes_list:
            modes_dict.update({f"l{ell}": {}})

            for emm in emm_list:
                message(f"Get modes dict l{ell} m{emm}", message_verbosity=3)
                value = self.mode(ell, emm)

                modes_dict[f"l{ell}"].update({f"m{emm}": value})

        return modes_dict

    @property
    def modes_list(self):
        message(f"ell_max of modes list {self.ell_max}", message_verbosity=3)
        message(f"spin weight {self.spin_weight}", message_verbosity=3)

        if not self._modes_list:
            self._modes_list = construct_mode_list(
                self.ell_max, self.spin_weight
            )

        # message(f"ell max of created modes list {max([item[0]
        # for item in self._modes_list])}",
        #        message_verbosity=4)

        return self._modes_list

    @property
    def ell_max(self):
        if self.Grid is not None:
            if self.Grid.grid_type == "GL":
                if self._ell_max is None:
                    self._ell_max = self.Grid.L

                assert self._ell_max <= self.Grid.L, (
                    "The grid L must be"
                    "greater than or equal to ell_max"
                    "of this SingleMode obj"
                )

        message(f"Accessing attr ell max {self._ell_max}", message_verbosity=4)

        return self._ell_max

    @property
    def spin_weight(self):
        """The spin weight of the modes"""
        return self._spin_weight

    def zero_modes(self, tol=1e-8):
        re_eval = False

        if not self._zero_modes:
            message("Zero modes will be computed afresh", message_verbosity=2)
            re_eval = True

        if self.tol is not None:
            if self.tol != tol:
                message(
                    "Tol has changed. Recalculating zero modes",
                    message_verbosity=2,
                )
                re_eval = True
                self._tol = tol

        if re_eval:
            self.compute_zero_modes(tol=self.tol)

        return self._zero_modes

    def non_zero_modes(self, tol=None):
        re_eval = False

        if not self._non_zero_modes:
            message("Zero modes will be computed afresh", message_verbosity=2)
            re_eval = True

        if tol is None:
            tol = self.tol

        if self.tol != tol:
            message(
                "Tol has changed. Recalculating zero modes",
                message_verbosity=2,
            )
            re_eval = True
            self._tol = tol

        if re_eval:
            self.compute_zero_modes(tol=tol)

        return self._non_zero_modes

    def mode(self, ell, emm):
        """Return a particular mode

        Parameters
        ----------
        ell : int
              The :math:`\\ell` index of the mode.
        emm : int
              The :math:`m` index of the mode.

        Returns
        -------
        mode_data : float
                    The value of the requested mode.
        """
        if abs(emm) > ell:
            raise ValueError(
                "Please request a valid mode ( abs(emm) > abs(ell) here)"
            )

        vec_idx = ell**2 + emm + ell

        return self._modes_data[vec_idx]

    def create_modes_array(self):
        """Create a modes array and initialize it with zeros.

        Parameters
        ----------
        ell_max : int
                  The maximum :math:`\\ell` value of the modes.

        Returns
        -------
        self.modes_array : modes_array
                           sets the `self.modes_array` attribute.

        Notes
        -----
        The ordering of axis is (ell, emm, extra axis)

        """

        message("Creating single mode data array", message_verbosity=3)

        ell_max = self.ell_max

        # Check ell_max
        if ell_max is None:
            try:
                ell_max = self.ell_max
            except Exception as ex:
                message(ex)
                raise NameError("Please supply ell_max")

        if self.modes_list is None:
            self._modes_list = construct_mode_list(
                ell_max=ell_max, spin_weight=self.spin_weight
            )

        if type(self.extra_mode_axes_shape) is tuple:
            self._modes_data = np.zeros(
                ((ell_max + 1) ** 2, *self.extra_mode_axes_shape),
                dtype=np.complex128,
            )
        else:
            self._modes_data = np.zeros(
                (ell_max + 1) ** 2, dtype=np.complex128
            )

    def construct_from_vec_modes(self, vec_modes):
        """Load the single modes object using the modes vector"""

        ell_max_vec_modes = int(np.sqrt(len(vec_modes)) - 1)

        if (ell_max_vec_modes + 1) ** 2 != len(vec_modes):
            raise ValueError(
                "The length of vec_modes does not match the filled modes array"
            )

        self._modes_data = vec_modes

    def set_mode_data(self, value, ell=None, emm=None):
        """Set the mode array data for the respective :math:`(\\ell, m)` mode.

        Parameters
        ----------
        ell : int
              The :math:`\\ell` polar mode number.
        emm_value : int
                    The :math:`m` azimuthal mode number.
        data : float
               The value of the mode data of the corresponding mode.

        """
        if (ell is None) and (emm is None):
            self._modes_data = np.array(value)
            # if len(value.shape)>1:
            #    self._extra_mode_axes_shape =
        else:
            # elif int(ell)==ell and int(emm)==emm:
            # Compute the linear vector index
            vec_idx = (ell) ** 2 + emm + ell
            message(f"Setting l{ell} m{emm} data {value}", message_verbosity=4)

            # Set the mode data.
            self._modes_data[vec_idx] = value
            message(
                f"Set mode data {self.mode(ell, emm)}", message_verbosity=4
            )
        # else:
        #    raise TypeError("Please provide integer values for ell and emm")

    def compute_zero_modes(self, tol=1e-8):
        """Get the details of the zero modes in
        the data below a given tolerance"""

        zero_modes_list = []
        non_zero_modes_list = []

        for ell, emm_list in self.modes_list:
            z_emm_list = []
            nz_emm_list = []

            for emm in emm_list:
                if abs(self.mode(ell, emm)) < tol:
                    z_emm_list.append(emm)
                else:
                    nz_emm_list.append(emm)

            if len(z_emm_list) > 0:
                zero_modes_list.append([ell, z_emm_list])

            if len(nz_emm_list) > 0:
                non_zero_modes_list.append([ell, nz_emm_list])

        self._zero_modes = zero_modes_list
        self._non_zero_modes = non_zero_modes_list

    def contract(
        self,
        ell_max=None,
        all_available_modes=True,
        only_modes_above_tol=False,
        tol=None,
    ):
        """Contract the modes to get co-ordinate space
        representation of the function.

        Parameters
        ----------
        ell_max: int, optional
                 The max :math:`\\ell` upto which to contract
                 the modes to. If None, then the available
                 `ell_max` will be chosen.

        all_available_modes: bool,optional
                             Wheter or not to use all available modes.

        Returns
        -------
        func: 2darray
              The function on the sphere in coordinate space
              representation.
        """

        if ell_max is None:
            ell_max = self.ell_max

        if self.Grid is not None:
            if self.Grid.grid_type == "GL":
                assert self.Grid.L >= ell_max, (
                    "The grid L must be greater than"
                    "or equal to ell_max of the requested SH expansion"
                )

        if all_available_modes:
            modes = self  # self.get_modes_dict()
        else:
            raise NotImplementedError

        if only_modes_above_tol:
            raise NotImplementedError
            modes = self.non_zero_modes(tol=tol)

        from spectral.spherical.transforms import SHContract
        from waveformtools.diagnostics import method_info

        minfo = method_info()

        func = SHContract(
            modes, self.Grid, ell_max, method_info=minfo, vectorize=True
        )

        return func

    def calculate_power_monitor_ell(self, ell_max=None):
        """Get the power in each ell mode

        Parameters
        ----------
        single_mode : modes dict / single_mode
                      A dictionary of modes or an instance
                      of the single mode class, that contains
                      the modes of an SH expansion.

        Returns
        -------
        power : list
                    A list containing the power
                    in each :math:`\\ell` mode
                    of the system.

        cumulative_power
        cumulative_even_power
        cumulative_even_power_axis
        cumulative_power_axis
        power_ell
        """

        if ell_max is None:
            ell_max = self.ell_max

        power = []

        for ell in range(ell_max + 1):
            power_this_ell = 0

            for emm in range(-ell, ell + 1):
                power_this_ell += np.absolute(self.mode(ell, emm)) ** 2

            power.append(power_this_ell)

        even_power = [
            power[index] + power[index + 1] for index in range(0, ell_max, 2)
        ]

        if ell_max % 2 == 1:
            even_power += [power[-1]]

        cumulative_even_power = [
            np.sum(even_power[:index]) for index in range(len(even_power))
        ]
        cumulative_power = [
            np.sum(power[:index]) for index in range(len(power))
        ]

        cumulative_even_power_axis = list(np.arange(0, ell_max, 2))

        if ell_max % 2 == 1:
            cumulative_even_power_axis += [ell_max]

        self.cumulative_power = cumulative_power
        self.cumulative_even_power = cumulative_even_power

        self.cumulative_even_power_axis = cumulative_even_power_axis
        self.power_axis = np.arange(ell_max + 1)

        self.power_ell = power

    def compare_modes(self, other_modes, prec=18):
        ell_max1 = self.ell_max
        ell_max2 = other_modes.ell_max

        ell_max = min(ell_max1, ell_max2)

        # result = False

        for ell in range(ell_max + 1):
            for emm in range(-ell, ell + 1):
                mode1 = other_modes.mode(ell, emm)
                mode2 = self.mode(ell, emm)

                np.testing.assert_almost_equal(
                    mode1, mode2, prec, f"The l{ell} m{emm} mode must equal"
                )

    def compare_modes_dict(self, other_modes_dict, prec=18):
        """Compare the Single modes object against a given modes
        dictionary"""

        for ell_key, ell_modes in other_modes_dict.items():
            ell = int(ell_key[1:])
            for emm_key, other_mode in ell_modes.items():
                emm = int(emm_key[1:])
                np.testing.assert_almost_equal(
                    self.mode(ell, emm),
                    other_mode,
                    prec,
                    f"The l{ell} m{emm} mode must equal at precision {prec}",
                )

    def truncate_modes(self, ell_max_choice):
        """Returns a new SingleModes object containing
        only modes upto the given `ell_max_choice`"""

        trunc_modes = SingleMode(
            ell_max=ell_max_choice,
            spin_weight=self.spin_weight,
            Grid=self.Grid,
            extra_mode_axes_shape=self.extra_mode_axes_shape,
            tol=self.tol,
        )

        trunc_modes._modes_data = self._modes_data[
            : ell_max_choice + 1, : 2 * ell_max_choice + 3
        ]

        return trunc_modes

    def get_expansion_residues(self, func=None):
        """Get the expansion residues"""

        if (np.array(func) == np.array(None)).all():
            func = self.func

        residues = [np.sum(func**2)]

        modes_list = construct_mode_list(ell_max=self.ell_max, spin_weight=0)
        # message(f"Modes list in SHContract
        # {modes_list}", message_verbosity=4)

        theta_grid, phi_grid = self.Grid.meshgrid

        recon_func = np.zeros(theta_grid.shape, dtype=np.complex128)

        for ell, emm_list in modes_list:
            for emm in emm_list:
                # Clm = modes[f"l{ell}"][f"m{emm}"]

                Clm = self.mode(ell, emm)
                message(
                    f"Clm shape in SHContract {Clm.shape}", message_verbosity=4
                )

                recon_func += Clm * Yslm_vec(
                    spin_weight=0,
                    ell=ell,
                    emm=emm,
                    theta_grid=theta_grid,
                    phi_grid=phi_grid,
                )

            res = np.sum((recon_func - func) ** 2)

            residues.append(res)

        return residues

    def get_in_spherepack_convention(self):
        """Convert the modes into spherepack convention. This
        returns a set of complex modes but only the real part
         is to be used. The imaginary part may be useful for
         error checking and should be small"""

        Rlm = []

        for emm in range(self.ell_max + 1):
            for ell in range(self.ell_max + 1):

                if abs(emm) > ell:
                    continue

                val_pm = self.mode(ell, emm)
                val_mm = self.mode(ell, -emm)

                if emm == 0:
                    factor = 1 / np.sqrt(2 * np.pi)
                    # val_sp*=np.sqrt(2*np.pi)
                    re_Ylm_val = val_pm.real

                else:
                    # val_sp*=np.sqrt(4*np.pi)
                    factor = 1 / np.sqrt(
                        4 * np.pi
                    )  # (-1)**emm * 5/2 #(-1)**emm * np.sqrt(4*np.pi)

                    re_Ylm_val = (
                        1 * (val_mm + (-1) ** emm * val_pm) / np.sqrt(2)
                    )

                re_Ylm_val *= factor

                Rlm.append(re_Ylm_val)

        self._modes_spherepack = Rlm

    def vector(self):
        """return the vector of modes in ell - emm format
        (opposite to spherepack's emm - ell format"""

        vec = self._modes_data.copy()
        # for ell in range(self.ell_max+1):
        #    for emm in range(-ell, ell+1):
        #        vec.append(self.mode(ell, emm))

        return np.array(vec)

    def evaluate_old(self, theta, phi, ell_max=None):
        """Evaluate the expansion at requested angular coordinates
         by looping over SWSH and summing. SWSHs are generated using
        `Yslm_vec`
        """

        if ell_max is None:
            ell_max = self.ell_max

        val = 0

        for ell in range(ell_max + 1):
            for emm in range(-ell, ell + 1):

                Ylm = Yslm_vec(
                    spin_weight=0,
                    theta_grid=theta,
                    phi_grid=phi,
                    ell=ell,
                    emm=emm,
                    cache=False,
                )

                val += self.mode(ell, emm) * Ylm

        # val = np.sum(self._modes_data * sYlm._modes_data)

        return val

    def evaluate_old_1(self, theta, phi, ell_max=None):
        """Evaluate the expansion at requested angular coordinates
        by generating SWSHs in serial and vectorizing the
        summation"""

        if ell_max is None:
            ell_max = self.ell_max

        from spectral.spherical.swsh import create_Yslm_modes_array

        Yslm = create_Yslm_modes_array(
            theta=float(theta),
            phi=float(phi),
            ell_max=ell_max,
            spin_weight=self.spin_weight,
        )

        val = np.sum(self._modes_data * Yslm.sYlm_modes._modes_data)

        return val

    def evaluate(self, theta=None, phi=None, ell_max=None):
        """Evaluate the expansion at requested angular coordinates
        by generating SWSHs in parallel and vectorizing the
        summation"""
        from spectral.spherical.Yslm_mp import Yslm_mp

        if ell_max is None:
            ell_max = self.ell_max

        if (np.array(theta) == np.array(None)).all() or (
            np.array(phi) == np.array(None)
        ).all():
            theta, phi = self.Grid.meshgrid

        sYlm = Yslm_mp(
            ell_max=ell_max, spin_weight=self.spin_weight, theta=theta, phi=phi
        )
        sYlm.run()
        Ylm_vec = sYlm.sYlm_modes._modes_data
        modes_data_len = len(Ylm_vec)

        val = np.tensordot(
            self._modes_data[:modes_data_len], Ylm_vec, axes=((0), (0))
        )

        return val

    def evaluate_sp(self, theta, phi, ell_max=None):
        """Evaluate the expansion at requested angular coordinates
        by computing SWSHs using the spherical package"""

        if ell_max is None:
            ell_max = self.ell_max

        from spectral.spherical.swsh import create_spherical_Yslm_modes_array

        sYlm = create_spherical_Yslm_modes_array(
            theta=float(theta),
            phi=float(phi),
            ell_max=ell_max,
            spin_weight=self.spin_weight,
        )

        val = np.sum(self._modes_data * sYlm._modes_data)

        return val

    def compute_spatial_detivatives(self):
        """Given the modes, compute its spatial derivatives"""

        assert self.spin_weight == 0, (
            "Derivatives have only been implemented"
            "for spin weight zero harmonics "
        )

        from qlmtools.differentiation import DerivSHFromSpec

        minfo = method_info(
            diff_method="SH", ell_max=self.ell_max, int_method="GL", reg=False
        )

        return DerivSHFromSpec(self, minfo)

    def plot_residues(self, orig_func=None):
        """Plot the residues of this expanion"""

        if (np.array(orig_func) == np.array(None)).all():
            orig_func = self.func

        residues = self.get_expansion_residues(orig_func)

        ell_axis = np.arange(-1, self.ell_max + 1)

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        ax.set_yscale("log")
        ax.scatter(ell_axis, residues, s=1)
        ax.set_title(r"Residues vs $\ell$")
        ax.set_ylabel("Residues")
        ax.set_xlabel(r"$\ell$")
        plt.show()



import os
import re
from pathlib import Path

import numpy as np

from waveformtools.sxs.prepare_waveforms import PrepareSXSWaveform
from waveformtools.waveformtools import (
    compute_chi_eff_from_masses_and_spins,
    compute_chi_prec_from_masses_and_spins,
    compute_masses_from_mass_ratio_and_total_mass,
    message,
)
import subprocess
import h5py
import pandas as pd


class SimulationExplorer:
    """Find and load simulations in a given directory.
    Use this to

    1. find available, running, and failed simulations,
    2. parse their parameters,
    3. compute secondary parameters
    4. export parameter tables to Markdown
    5. Keep a track of already processed waveforms
    6. process waveforms in batch if not i.r. extrapolate and CoM correct

    Attributes
    ----------
    search_dir: str/Path
                The directory containing simulations
    available_sim_names: list of str
                     All discovered simulation names
    available_sim_paths: dict of str/Path
                     The full paths of all found simulations
    highest_ecc_nums: dict of str/Path
                      The highest available Ecc dir across
                      all simulations
    all_sim_params: dict
                    All parsed and available parameters
                    across all discovered simulations.
    all_sim_params_df: DataFrame
                       Same as `all_sim_params` but a sorted data frame
    prepared_waveforms: list of str
                        Simulations whose waveforms had already
                        been processed, that were found in the
                        `prepared_waveforms_dir`
    available_sim_levels: dict of int
                          The available levels within each simulation.
    waveforms_to_prepare: list of str
                          The waveforms that are waiting to be processed.
    processed_waveforms: list of str
                         Simulations whose waveforms have been processed
                         using methods in this class.
    prepared_waveforms_dir: str/Path
                            The full path to the directory
                            where all processed waveforms are
                            being saved.
    bfi_home_dir: str/Path
                  The full path to the BFI home dir
    bfi_project_name: str
                      The name of the BFI project
    bfi_sim_params: dict
                    The simulation parameters as defined in
                    the bfi Project
    inspiral_segments, ringdown_segments: dict
                                          The inspiral and ringdown segements
                                          of all available simulations
    sim_status: dict
                The current status of all sims

    Methods
    -------
    check_for_ecc_dir
        Check if  given directory is an Ecc directory
    search_simulations
        Search for simulations
    find_highest_ecc
        Given the full path to a sim,
        find the highest available Ecc directory
    discover_levels
        find all available levels given the full path
        to a simulation
    get_all_segments
        Given s sim, find all available segments
        across all available levs
    concatenate_ascii_data
        given a sim, lev and a file name, concatenate
        the file across segments.
    get_ecc_dirs
        get all ecc dirs given the full path to a simulation
    find_highest_ecc_dirs
        Discover the highest ecc dirs across all available simulations
    parse_sim_params_input_file
        Load parameters of all the discovered simulations from their
        Params.input file
    parse_sim_target_params_input_file
        Load parameters of all the discovered simulations from their
        TargetParams.input file
    fetch_sim_params
        Fetch the simulation parameters across the discovered
        simulations.
    strip_sim_name_from_waveform_dir
        Given the name of a directory in `prepared_waveforms_dir`,
        parse the simulation name.
    discover_extrapolated_sims
        Given the already-prepared waveforms in `prepared_waveforms_dir`,
        discover the corresponding simulations.
    get_sim_names_from_prepared_waveform_dirs
        Given a list of directories in `prepared_waveforms_dir`,
        get their corresponding simulation names.
    prepare_list_of_sims
        Find simulations that need waveform processing.
    prepare_waveforms
        Process the waveforms
    compute_masses
        COmpute the masses of all BHs from their mass ratios.
    compute_chi_eff
        Compute the :math:`\\chi_{eff}` parameters of all the
        simulations
    compute_chi_prec
        Compute the :math:`\\chi_{prec}` parameters of all the
        simulations
    get_all_chi_eff
        Fetch the :math:`\\chi_{eff}` parameter of all simulations
    get_all_chi_prec
        Fetch the :math:`\\chi_{prec}` parameter of all simulations
    get_all_mass_ratios
        Fetch the mass-ratio :math:`q` parameter of all simulations
    write_history
        Write processed waveforms in this session to a history file.
    """

    def __init__(
        self,
        search_dir="./",
        prepared_waveforms_dir=None,
        bfi_project_name=None,
        bfi_home_dir=None,
        ref_time=None,
        filter_items=[
            "Sl. No",
            "Status",
            "ReqRefTime",
            "Omega0",
            "Omega_ref",
            "Ncycles",
            "Ncycles (2,2)",
            "D0",
            "MassRatio",
            "ChiA",
            "ChiA_ref",
            "ChiB",
            "ChiB_ref",
            "ChiEff",
            "ChiPrec",
        ],
    ):
        self._search_dir = Path(search_dir)

        self._prepared_waveforms_dir = Path(prepared_waveforms_dir)
        self._ref_time = ref_time
        self._filter_items = filter_items

        self._bfi_project_name = bfi_project_name
        self._bfi_home_dir = Path(bfi_home_dir)
        self._sim_basename = None
        self._bfi_sim_params = {}
        self._all_req_ref_times = {}
        self._all_req_ref_time_segments = []

    @property
    def search_dir(self):
        return self._search_dir

    @property
    def filter_items(self):
        return self._filter_items

    @property
    def sim_basename(self):
        return self._sim_basename

    @property
    def bfi_project_name(self):
        return self._bfi_project_name

    @property
    def bfi_home_dir(self):
        return self._bfi_home_dir

    @property
    def bfi_sim_params(self):
        return self._bfi_sim_params

    @property
    def available_sim_names(self):
        return self._available_sim_names

    @property
    def available_sim_paths(self):
        return self._available_sim_paths

    @property
    def highest_ecc_nums(self):
        return self._highest_ecc_nums

    @property
    def all_sim_params(self):
        return self._all_sim_params

    @property
    def all_sim_params_df(self):
        return self._all_sim_params_df

    @property
    def prepared_waveforms(self):
        return self._prepared_waveforms

    @property
    def available_sim_levs(self):
        return self._available_sim_levs

    @property
    def processed_waveforms(self):
        return self._processed_waveforms

    @property
    def prepared_waveforms_dir(self):
        return self._prepared_waveforms_dir

    @property
    def waveforms_to_prepare(self):
        return self._waveforms_to_prepare

    @property
    def inspiral_segments(self):
        return self._inspiral_segments

    @property
    def ringdown_segments(self):
        return self._ringdown_segments

    @property
    def sim_status(self):
        return self._sim_status

    @property
    def ncycles(self):
        return self._ncycles

    @property
    def ncycles_wf(self):
        return self._ncycles_wf

    @property
    def ref_time(self):
        return self._ref_time

    @property
    def all_req_ref_times(self):
        return self._all_req_ref_times

    @property
    def all_req_ref_time_segments(self):
        return self._all_req_ref_time_segments

    @property
    def deleted_sims(self):
        return self._deleted_sims

    def check_for_ecc_dir(self, path_of_dir_to_check):
        """Check if the given directory is an Ecc directory"""

        flag = True

        message(f"Checking {path_of_dir_to_check}", message_verbosity=3)

        if not os.path.isdir(path_of_dir_to_check):
            flag = False
            message(
                f"{path_of_dir_to_check} is not a directory",
                message_verbosity=3,
            )

            return flag

        subsubdirs = os.listdir(path_of_dir_to_check)

        if "Ecc0" not in subsubdirs:
            flag = False
            message(
                f"{path_of_dir_to_check} does not contain Ecc0",
                message_verbosity=3,
            )
            return flag

        ecc0_dir = os.listdir(path_of_dir_to_check.joinpath("Ecc0"))

        if "DoMultipleRuns.input" not in ecc0_dir:
            flag = False
            message(
                f"{path_of_dir_to_check} does not contain"
                "DoMultipleRuns.input",
                message_verbosity=3,
            )

        if flag:
            message(
                f"\t {path_of_dir_to_check} is a simulation directory",
                message_verbosity=3,
            )

        return flag

    def search_simulations(self):
        """Search for simulations within the search directories"""

        available_sim_paths = {}
        available_sim_names = []

        one_search_dir = self.search_dir

        message(f"Searching in {one_search_dir}", message_verbosity=1)

        subdirs = os.listdir(one_search_dir)

        message(f"All subdirs {subdirs}", message_verbosity=2)

        for possible_sim_name in subdirs:
            possible_sim_path = one_search_dir.joinpath(possible_sim_name)

            message(f"\tSearching in {possible_sim_path}", message_verbosity=2)

            flag = self.check_for_ecc_dir(possible_sim_path)

            if flag:
                available_sim_paths.update(
                    {possible_sim_name: possible_sim_path}
                )

                available_sim_names.append(possible_sim_name)

        self._available_sim_names = available_sim_names
        self._available_sim_paths = available_sim_paths

        self.compute_sim_basename()
        self.find_highest_ecc_dirs()

    def find_highest_ecc(self, sim_path):
        """Find the highest ecc level in a given sim dir"""

        ecc_dirs = self.get_ecc_dirs(sim_path)

        available_ecc_nums = [item.strip("Ecc") for item in ecc_dirs]

        available_ecc_nums = [
            int(item) for item in available_ecc_nums if item.isdigit()
        ]

        # print(available_eccs)

        highest_ecc_num = np.argmax(available_ecc_nums)

        return highest_ecc_num

    def discover_levels_old(self):
        """Discover all avaiable levels across all simulations"""

        all_levs = {}

        for sim_name, sim_path in self.available_sim_paths.items():
            sim_levs = []

            path_to_highest_ecc_dir = sim_path.joinpath(
                f"Ecc{self.highest_ecc_nums[sim_name]}/Ev"
            )

            subdirs = os.listdir(path_to_highest_ecc_dir)

            subdirs = [item for item in subdirs if "Ringdown" in item]

            if subdirs is None:
                continue

            levs = [item for item in subdirs if "Lev" in item]
            levs = [item for item in levs if "SysInit" not in item]

            for item in levs:
                result = re.search("Lev[0-9]_Ringdown", item)

                lev_num = int(result.group()[3:4])

                sim_levs.append(lev_num)

            sim_levs = list(set(sim_levs))

            all_levs.update({sim_name: sim_levs})

        self._available_sim_levs = all_levs

    def discover_levels(self):
        """Discover all avaiable levels across all simulations"""

        all_levs = {}

        for sim_name in self.available_sim_names:

            sim_levs = []

            for sim_lev in self.inspiral_segments[sim_name].keys():

                this_lev = int(sim_lev[3:])

                sim_levs.append(this_lev)

            all_levs.update({sim_name: sim_levs})

        self._available_sim_levs = all_levs

    def discover_segments(self):
        """Find all the segments across all simulations"""
        import os

        all_inspiral_segments = {}
        all_ringdown_segments = {}

        for sim_name in self.available_sim_names:

            message(f"Loading segments in {sim_name}", message_verbosity=2)
            ecc_ev_dir = self.available_sim_paths[sim_name].joinpath(
                f"Ecc{self.highest_ecc_nums[sim_name]}/Ev"
            )

            inspiral_segments, ringdown_segments = (
                self.get_all_one_sim_segments(ecc_ev_dir)
            )

            all_inspiral_segments.update({sim_name: inspiral_segments})

            all_ringdown_segments.update({sim_name: ringdown_segments})

        self._inspiral_segments = all_inspiral_segments
        self._ringdown_segments = all_ringdown_segments

    def get_all_one_sim_segments(self, ecc_ev_dir):
        """Get the segments within one Ecc dir"""
        import os

        inspiral_segments = {}
        ringdown_segments = {}

        available_dirs = os.listdir(ecc_ev_dir)
        available_dirs = [
            item for item in available_dirs if ".bak" not in item
        ]

        message("Available dirs", available_dirs, message_verbosity=2)

        available_lev_dirs = [item for item in available_dirs if "Lev" in item]
        available_lev_dirs = [
            item for item in available_lev_dirs if "SysInit" not in item
        ]

        available_lev_dirs = [
            item for item in available_lev_dirs if "Ringdown" not in item
        ]
        available_lev_dirs = [
            item
            for item in available_lev_dirs
            if re.search("Lev[0-9]_[A-Z]+", item) is not None
        ]

        message("Available lev dirs", available_lev_dirs, message_verbosity=2)

        available_ringdown_lev_dirs = [
            item for item in available_dirs if "_Ringdown" in item
        ]
        available_ringdown_lev_dirs = [
            item
            for item in available_ringdown_lev_dirs
            if re.search("Lev[0-9]_Ringdown", item) is not None
        ]
        message(
            "Available ringdown lev dirs",
            available_ringdown_lev_dirs,
            message_verbosity=2,
        )

        available_levs = set([int(item[3:-3]) for item in available_lev_dirs])
        message("Available levs", available_levs, message_verbosity=2)

        available_ringdown_levs = set(
            [int(item[3:-9]) for item in available_ringdown_lev_dirs]
        )
        message(
            "Available ringdown levs",
            available_ringdown_levs,
            message_verbosity=2,
        )

        # Fetch segments
        for lev in available_levs:

            this_lev_segments = [
                item for item in available_lev_dirs if f"Lev{lev}" in item
            ]
            this_lev_segments = sorted(
                [item[5:] for item in this_lev_segments]
            )
            inspiral_segments.update({f"Lev{lev}": this_lev_segments})

            if lev in available_ringdown_levs:
                inside_ringdown_dirs = os.listdir(
                    ecc_ev_dir.joinpath(f"Lev{lev}_Ringdown")
                )
                inside_ringdown_segments = [
                    item for item in inside_ringdown_dirs if "Lev" in item
                ]
                inside_ringdown_segments = [
                    item
                    for item in inside_ringdown_segments
                    if "SysInit" not in item
                ]
                inside_ringdown_segments = [
                    item
                    for item in inside_ringdown_segments
                    if re.search("Lev[0-9]_[A-Z]+", item) is not None
                ]
                this_lev_ringdown_segments = sorted(
                    [item[5:] for item in inside_ringdown_segments]
                )
                ringdown_segments.update(
                    {f"Lev{lev}": this_lev_ringdown_segments}
                )

        return inspiral_segments, ringdown_segments

    def concatenate_ascii_data(self, fname, ecc, lev, sim_dir):
        """Concatenate the data in the given file name, lev,
        ecc dir"""

        full_ecc_path = f"{sim_dir}/Ecc{ecc}/Ev/"

        # get segments
        all_super_segments = self.get_all_segments(full_ecc_path)

        all_segments = all_super_segments[f"Lev{lev}"]

        all_data = []

        for item in all_segments:
            file = f"{full_ecc_path}/Lev{lev}_{item}/Run/{fname}"

            one_segment_data = np.genfromtxt(file)
            all_data.append(one_segment_data)

        all_data = np.concatenate((all_data))

        return all_data

    def get_ecc_dirs(self, sim_path):
        """Get all ecc directories within a sim dir"""
        subdirs = os.listdir(sim_path)

        ecc_dirs = [item for item in subdirs if "Ecc" in item]

        return ecc_dirs

    def find_highest_ecc_dirs(self):
        """Find and load the highest ecc level available"""

        highest_ecc_nums = {}

        for sim_name, sim_path in self.available_sim_paths.items():
            highest_ecc_num = self.find_highest_ecc(sim_path)

            highest_ecc_nums.update({sim_name: highest_ecc_num})

        self._highest_ecc_nums = highest_ecc_nums

    def compute_sim_basename(self):
        """Compute the sim basename from the dir names"""

        one_sim_name = self.available_sim_names[-1]

        r1 = re.search("[a-zA-Z][0-9]+", one_sim_name)

        end_ind = r1.start() + 1

        self._sim_basename = one_sim_name[:end_ind]

    def parse_sim_params_input_file(self, path_to_params_input_file):
        """Parse all sim params given path to the Params.input file"""

        sim_params = {}

        message(
            "Parsing simulation parameters.input file"
            f"for {path_to_params_input_file}",
            message_verbosity=2,
        )

        with open(path_to_params_input_file, "r") as pfile:
            eof = False

            while not eof:
                line = pfile.readline()

                if "Omega0" in line:
                    val = re_fetch_float(line)

                    sim_params.update({"Omega0": val})

                if "adot0" in line:
                    val = re_fetch_float(line)

                    sim_params.update({"adot0": val})

                if "D0" in line:
                    val = re_fetch_float(line)

                    sim_params.update({"D0": val})

                if "MassRatio" in line:
                    val = re_fetch_float(line)

                    sim_params.update({"MassRatio": val})

                if "SpinA" in line:
                    val = re_fetch_vector(line)

                    sim_params.update({"ChiA": val})

                if "SpinB" in line:
                    val = re_fetch_vector(line)

                    sim_params.update({"ChiB": val})

                if "$IDType" in line:
                    val = re_fetch_string(line)

                    sim_params.update({"IDType": val})

                    eof = True

        return sim_params

    def parse_sim_target_params_input_file(
        self, path_to_target_params_input_file
    ):
        """Parse a TargetParams.input file given a sim"""

        if not os.path.exists(path_to_target_params_input_file):
            return None

        else:
            target_params = {}

            with open(path_to_target_params_input_file, "r") as tpfile:
                eof = False

                while not eof:
                    line = tpfile.readline()

                    if "$ReferenceTime" in line:
                        val = re_fetch_float(line)

                        target_params.update({"ReferenceTime": val})

                    if "$SemiMajorAxis" in line:
                        val = re_fetch_float(line)

                        target_params.update({"SemiMajorAxis": val})

                    if "$Eccentricity" in line:
                        val = re_fetch_float(line)

                        target_params.update({"Eccentricity": val})

                    if "$AnomalyAngle" in line:
                        val = re_fetch_float(line)

                        target_params.update({"AnomalyAngle": val})

                    if "$IDType" in line:
                        eof = True

            return target_params

    def parse_bfi_params_file(self):
        """Parse parameters from a BFI params file"""

        bfi_sim_params = {}

        bfi_sim_basename = self.bfi_project_name

        if self.sim_basename != bfi_sim_basename:

            message(
                f"sim base name: {self.sim_basename} \t"
                f"bfi project name: {bfi_sim_basename}"
            )

            raise KeyError(
                "The BFI project name "
                "do not match sim basenames found in search dir"
            )

        else:
            projects_rel_dir = Path(
                f"ParamFiles/{self.bfi_project_name}/params"
            )
            bfi_params_file = self.bfi_home_dir.joinpath(projects_rel_dir)

            with open(bfi_params_file, "r", encoding="utf-8") as pf:
                for line in pf:
                    if line is not None:
                        # Get params
                        results = re.findall(
                            "[a-zA-Z]+=[0-9]+[.]?[0-9]*", line
                        )
                        fresults = [
                            item for item in results if "chi" not in item
                        ]
                        fresults = [
                            item for item in fresults if "q" not in item
                        ]

                        if len(fresults) == 0:
                            continue

                        message("fresults", fresults, message_verbosity=3)

                        # Assuming that ID is the first element
                        sim_name = (
                            self.sim_basename + fresults[0].split("=")[1]
                        )

                        bfi_sim_params.update({sim_name: {}})

                        for item in fresults[1:]:

                            key, value = item.split("=")
                            value = float(value)

                            bfi_sim_params[sim_name].update({key: value})
        self._bfi_sim_params = bfi_sim_params

        # Delete empty sims from bfi file
        for item in self.deleted_sims:

            self._bfi_sim_params.pop(item)

        self.append_bfi_sim_params_to_all_params()

    def append_bfi_sim_params_to_all_params(self):
        """Append the contents of the BFI sim params to the master
        params dict"""

        for sim_name, sim_params in self.bfi_sim_params.items():
            # for sim_name in self.available_sim_names:

            sim_params = self.bfi_sim_params[sim_name]

            available_sim_param_keys = self.all_sim_params[sim_name].keys()

            for key, value in sim_params.items():

                if key not in available_sim_param_keys:

                    self._all_sim_params[sim_name].update({key: value})

    def fetch_sim_params(self):
        """Get all the simulations' parameters"""

        all_sim_params = {}

        for sim_name, sim_path in self.available_sim_paths.items():
            ecc_num = self.highest_ecc_nums[sim_name]

            path_to_params_input_file = sim_path.joinpath(
                f"Ecc{ecc_num}"
            ).joinpath("Params.input")
            path_to_target_params_input_file = sim_path.joinpath(
                f"Ecc{ecc_num}"
            ).joinpath("TargetParams.input")

            one_sim_params_dict = self.parse_sim_params_input_file(
                path_to_params_input_file
            )

            one_sim_target_params_dict = (
                self.parse_sim_target_params_input_file(
                    path_to_target_params_input_file
                )
            )

            one_sim_params = one_sim_params_dict

            if one_sim_target_params_dict is not None:
                one_sim_params.update(one_sim_target_params_dict)

            all_sim_params.update({sim_name: one_sim_params})

        self._all_sim_params = all_sim_params
        self.parse_bfi_params_file()
        self.compute_chi_eff()
        self.compute_chi_prec()

        self.compute_ncycles()
        self.compute_ncycles_from_waveform()

        self.prepare_pandas_dataframe()

    def prepare_pandas_dataframe(self):
        """Prepare a pandas dataframe from the sim params
        found"""

        import pandas as pd

        all_params_df = pd.DataFrame(self.all_sim_params)
        all_params_df = all_params_df.transpose()
        all_params_df = all_params_df.sort_index()

        all_params_df.insert(0, "Sl. No.", range(1, 1 + len(all_params_df)))
        # self.discover_segments()
        # self.discover_levels()
        all_params_df.insert(1, "Status", self.get_basic_sim_status())
        all_params_df.insert(3, "Ncycles", self.ncycles)
        all_params_df.insert(4, "Ncycles (2,2)", self.ncycles_wf)

        self._all_sim_params_df = all_params_df

    def get_basic_sim_status(self):
        """Insert status of individual sim levs
        into the dataframe"""

        basic_sim_status_dict = {}

        for sim_name_with_lev, lev_status in self.sim_status["Status"].items():
            sim_name, lev_name = sim_name_with_lev.split("_")
            lev_num = int(lev_name[3:])

            if not sim_name in basic_sim_status_dict.keys():
                basic_sim_status_dict.update({sim_name: {lev_num: lev_status}})
            else:
                basic_sim_status_dict[sim_name].update({lev_num: lev_status})

        return basic_sim_status_dict

    def update_pandas_dataframe(self):
        import pandas as pd

        self._all_sim_params_df = pd.DataFrame(
            self.all_sim_params
        ).T.sort_index()

        self.prepare_pandas_dataframe()

    def strip_sim_name_from_waveform_dir(self, wdir):
        """Extract the sim name from its processed waveform
        dir name"""
        result = re.search("[A-Za-z]+[0-9]+_", wdir)

        if result is not None:
            return result.group()[:-1]

        else:
            return None

    def discover_extrapolated_sims(self):
        """Read in all available waveform directories
        in the given directory and create a list of sims
        whose waveforms have already been prepared"""

        waveforms_dir = self.prepared_waveforms_dir

        all_wfs = os.listdir(waveforms_dir)

        message("Contents of waveforms dir", message_verbosity=2)

        all_wfs = [
            item
            for item in all_wfs
            if os.path.isdir(waveforms_dir.joinpath(item))
        ]

        message("Dirs of waveforms dir", message_verbosity=2)

        # Strip suffix
        sim_names = self.get_sim_names_from_prepared_waveform_dirs(all_wfs)

        message("sims in waveforms dir", message_verbosity=2)
        self._prepared_waveforms = sim_names

    def get_sim_names_from_prepared_waveform_dirs(
        self, prepared_waveform_dirs
    ):
        """Get all sim names from prepared waveform dirs"""
        prepared_waveform_sim_names = []

        for item in prepared_waveform_dirs:
            one_sim_name = self.strip_sim_name_from_waveform_dir(item)

            if one_sim_name is not None:
                prepared_waveform_sim_names.append(one_sim_name)

        prepared_waveform_sim_names = list(set(prepared_waveform_sim_names))

        return prepared_waveform_sim_names

    def prepare_list_of_sims(self):
        """Prepare a non-redundant list of simulations
        to run waveform preps for"""

        self.discover_extrapolated_sims()

        waveforms_to_prep = []

        for sim_name in self.available_sim_names:
            if (sim_name not in self.prepared_waveforms) and (
                self.available_sim_levs[sim_name] is not None
            ):
                waveforms_to_prep.append(sim_name)

        self._waveforms_to_prepare = waveforms_to_prep

    def prepare_waveforms(self, simulations_to_prepare=None):
        """Prepare waveforms that havent been prepared yet
        by extrapolating, CoM correcting"""

        self.prepare_list_of_sims()

        processed_waveforms = {}

        if (np.array(simulations_to_prepare) == np.array(None)).any():
            simulations_to_prepare = self.waveforms_to_prepare

        else:
            simulations_to_prepare = [
                item
                for item in simulations_to_prepare
                if item not in self.prepared_waveforms
            ]

        for sim_name in simulations_to_prepare:
            processed_levs = []

            available_sim_levs = self.available_sim_levs[sim_name]

            ecc = [self.highest_ecc_nums[sim_name]]

            success_sims = []

            for lev in available_sim_levs:
                message(
                    "Preparing waveform for sim ",
                    sim_name,
                    "Lev ",
                    lev,
                    message_verbosity=2,
                )

                try:
                    wfp = PrepareSXSWaveform(
                        sim_name=sim_name,
                        sim_dir=self.available_sim_paths[sim_name],
                        lev=lev,
                        out_dir=self.prepared_waveforms_dir
                        / f"{sim_name}_waveforms_Lev{lev}",
                    )

                    flag = wfp.prepare_waveform()

                    if flag:
                        processed_levs.append(lev)

                except Exception as excep:
                    message(excep)
                    message(
                        "Waveform processing failed sim"
                        f"{sim_name}, lev {lev}"
                    )

            processed_waveforms.update({sim_name: processed_levs})

        self._processed_waveforms = processed_waveforms

    def compute_masses(self):
        """Compute the individual masses of all sims"""
        for sim_name in self.available_sim_names:
            mass1, mass2 = compute_masses_from_mass_ratio_and_total_mass(
                self.all_sim_params[sim_name]["MassRatio"]
            )

            self.all_sim_params[sim_name].update(
                {"IndividualMasses": (mass1, mass2)}
            )

    def compute_chi_eff(self):
        """Compute the :math:`\\chi_{eff} parameter` across all sims"""
        for sim_name in self.available_sim_names:
            mass_ratio = self.all_sim_params[sim_name]["MassRatio"]

            spin1 = self.all_sim_params[sim_name]["ChiA"]
            spin2 = self.all_sim_params[sim_name]["ChiB"]

            chi_eff = compute_chi_eff_from_masses_and_spins(
                spin1, spin2, mass_ratio
            )

            self.all_sim_params[sim_name].update({"ChiEff": chi_eff})

    def compute_chi_prec(self):
        """Compute the :math:`\\chi_{prec}` parameters across all sims"""
        for sim_name in self.available_sim_names:
            mass_ratio = self.all_sim_params[sim_name]["MassRatio"]

            spin1 = self.all_sim_params[sim_name]["ChiA"]
            spin2 = self.all_sim_params[sim_name]["ChiB"]

            chi_prec = compute_chi_prec_from_masses_and_spins(
                spin1, spin2, mass_ratio
            )

            self.all_sim_params[sim_name].update({"ChiPrec": chi_prec})

    def get_all_chi_eff(self):
        """Fetch all :math:`\\chi_{eff}` parameters"""
        all_chi_eff = []

        for sim_name in self.available_sim_names:
            all_chi_eff.append(self.all_sim_params[sim_name]["ChiEff"])

        return all_chi_eff

    def get_all_mass_ratios(self):
        """Fetch all mass-ratios"""

        all_mass_ratios = []

        for sim_name in self.available_sim_names:
            all_mass_ratios.append(self.all_sim_params[sim_name]["MassRatio"])

        return all_mass_ratios

    def write_history(self):
        raise NotImplementedError

        with open(self.history_file, "a") as th:
            for item in dirs:
                th.write(item)
                th.write("\n")

    def get_last_segment(self, sim_name, lev):
        """Get the last available segment of a sim"""

        message(
            f"Fetching last segment for sim {sim_name}, lev{lev}",
            message_verbosity=2,
        )

        RINGDOWN_INITIATED = False

        ringdown_segments = self.ringdown_segments[sim_name]

        if ringdown_segments:

            lev_key = f"Lev{lev}"

            if (
                lev_key in ringdown_segments.keys()
                and len(ringdown_segments[lev_key]) > 0
            ):
                RINGDOWN_INITIATED = True
                last_segment = sorted(ringdown_segments[lev_key])[-1]
                last_segment_path = self.available_sim_paths[
                    sim_name
                ].joinpath(
                    f"Ecc{self.highest_ecc_nums[sim_name]}"
                    f"/Ev/Lev{lev}_Ringdown/Lev{lev}_{last_segment}"
                )

            else:
                RINGDOWN_INITIATED = False

        if RINGDOWN_INITIATED is False:
            inspiral_segments = self.inspiral_segments[sim_name]

            if inspiral_segments:
                last_segment = sorted(inspiral_segments[f"Lev{lev}"])[-1]
                last_segment_path = self.available_sim_paths[
                    sim_name
                ].joinpath(
                    f"Ecc{self.highest_ecc_nums[sim_name]}"
                    f"/Ev/Lev{lev}_{last_segment}"
                )

            else:
                raise KeyError(f"No segments in sim {sim_name}")

        return last_segment, last_segment_path, RINGDOWN_INITIATED

    def get_segment_jobid(self, last_segment_path):
        # Check if segment is running
        # Get Jobid

        all_jobids = []

        spec_jobid_file = last_segment_path.joinpath("SpEC.jobid")

        if not os.path.exists(spec_jobid_file):
            return -1

        with open(spec_jobid_file, "r", encoding="utf-8") as sjf:
            for line in sjf:
                one_jobid = line.split()[0]
                all_jobids.append(int(one_jobid))

        latest_jobid = all_jobids[-1]
        return latest_jobid

    def check_if_job_is_running(self, jobid):
        """Check if a segment is runnning"""

        message(
            f"Checking if the job {jobid} is running...", message_verbosity=2
        )

        # Get running jobs
        cmd_stdout = subprocess.run(
            ["squeue -u ${USER} | awk '{print $1}'"],
            shell=True,
            stdout=subprocess.PIPE,
        ).stdout

        cmd_stdout_status = subprocess.run(
            ["squeue -u ${USER} | awk '{print $5}'"],
            shell=True,
            stdout=subprocess.PIPE,
        ).stdout

        cmd_stdout_nodes = subprocess.run(
            ["squeue -u ${USER} | awk '{print $8}'"],
            shell=True,
            stdout=subprocess.PIPE,
        ).stdout

        running_jobs = cmd_stdout.decode().split("\n")[1:-1]
        running_jobs = [int(item) for item in running_jobs]

        # Get status
        jobs_status = cmd_stdout_status.decode().split("\n")[1:-1]

        nodes = cmd_stdout_nodes.decode().split("\n")[1:-1]

        jobs = {}

        for index in range(len(running_jobs)):
            one_jobid = running_jobs[index]
            one_status = jobs_status[index]
            one_node = nodes[index]

            jobs.update({one_jobid: {"Status": one_status, "Node": one_node}})

        # jobs = dict(zip(running_jobs, jobs_status))

        message("Running jobs ", running_jobs, message_verbosity=2)
        message("sim job id", jobid, message_verbosity=2)

        if jobid in running_jobs:
            status = jobs[jobid]["Status"]
            message("Segment running", message_verbosity=2)

            if status == "R":
                SEGMENT_STATUS = "Running"
            elif status == "PD":
                SEGMENT_STATUS = "Queued"
            else:
                SEGMENT_STATUS = "Unknown"
        else:
            SEGMENT_STATUS = "Not Running"

        try:
            jobs[jobid]["Status"] = SEGMENT_STATUS
            return jobs[jobid]
        except:
            return None

    def check_for_errors(self, last_segment_path):
        """Check for error files to learn about the status"""

        # Check for errors
        run_dir = last_segment_path.joinpath("Run")

        files_in_run_dir = os.listdir(run_dir)

        error_files = [
            item for item in files_in_run_dir if "Errors000" in item
        ]

        message("Error files", error_files, message_verbosity=2)

        if len(error_files) > 0:
            SEGMENT_ERROR = True
        else:
            SEGMENT_ERROR = False
            message("Sim status is unknown", message_verbosity=2)
            status = "Unknown"

        return SEGMENT_ERROR

    def read_last_segment_status(self, sim_name, lev):
        """Find out the status of the last segment of a particular sim"""

        one_sim_status = {}

        last_segment, last_segment_path, RINGDOWN_INITIATED = (
            self.get_last_segment(sim_name, lev)
        )

        if RINGDOWN_INITIATED:
            last_segment = "RD_" + last_segment

        sim_key = f"{sim_name}_Lev{lev}"
        one_sim_status.update({sim_key: {"Segment": last_segment}})

        SIM_COMPLETE = False
        SEGMENT_ERROR = False
        status = None
        SEGMENT_TO_SUBMIT = False
        # SEGMENT_RUNNING=False
        allocation = None
        # Check not submitted
        spec_run_dir = last_segment_path.joinpath("Run")
        spec_out_file = last_segment_path.joinpath("Run/SpEC.out")

        if not os.path.exists(spec_run_dir):
            SEGMENT_COMPLETE = False
            SIM_COMPLETE = False
            # SEGMENT_RUNNING = SEGMENT_RUNNING_STATUS
            SEGMENT_TO_SUBMIT = True

            SEGMENT_RUNNING_STATUS = "Not Running"
            status = "To be submitted"
            # else:
            # status = SEGMENT_RUNNING_STATUS
            segment_status_comment = "No Run dir"
        elif not os.path.exists(spec_out_file):
            SEGMENT_COMPLETE = False
            SIM_COMPLETE = False
            SEGMENT_RUNNING = "Not Running"

            segment_status_comment = "SpEC.out not found"
            status = "Halted"

        # Check if terminated
        else:

            segment_jobid = self.get_segment_jobid(last_segment_path)

            if segment_jobid < 0:
                SEGMENT_RUNNING_STATUS = "Not Running"
            else:
                seg_status = self.check_if_job_is_running(segment_jobid)

                if seg_status is not None:
                    SEGMENT_RUNNING_STATUS = seg_status["Status"]
                    allocation = seg_status["Node"]
                else:
                    SEGMENT_RUNNING_STATUS = "Not Running"

            with open(spec_out_file, "r", encoding="utf-8") as sof:
                lines = sof.readlines()[-3:]

                message("Lines from SpEC.out", lines, message_verbosity=2)

                message(
                    "Segment runnint 2",
                    SEGMENT_RUNNING_STATUS,
                    message_verbosity=2,
                )
                SEGMENT_ERROR = self.check_for_errors(last_segment_path)

                if "To be continued...\n" in lines:
                    SEGMENT_COMPLETE = True
                    SIM_COMPLETE = False
                    assert (
                        SEGMENT_RUNNING_STATUS == "Not Running"
                    ), "Segment was found to be completed. It cant be running!"
                    assert (
                        SEGMENT_ERROR == False
                    ), "Segment was found to have errors but also complete!"

                    segment_status_comment = [
                        item
                        for item in lines
                        if "Termination condition" in item
                    ][-1]

                    status = "Halted"

                elif "Termination condition FinalTime\n" in lines:
                    SEGMENT_COMPLETE = True
                    SIM_COMPLETE = True
                    assert (
                        SEGMENT_RUNNING_STATUS == "Not Running"
                    ), "Segment was found to be completed. It cant be running!"

                    # assert SEGMENT_ERROR == False, "Segment was found
                    # to have errors but also complete!"
                    if SEGMENT_ERROR == True:
                        message(
                            "Segment was found to have errors but also complete!"
                        )

                    segment_status_comment = "Termination condition FinalTime"
                    status = "Completed"

                elif "t=" in lines[-1]:
                    SEGMENT_COMPLETE = False
                    SIM_COMPLETE = False

                    if SEGMENT_RUNNING_STATUS == "Running":
                        status = "Running"
                    elif SEGMENT_ERROR:
                        status = "Error"
                    else:
                        status = "Halted"

                    segment_status_comment = status
                else:
                    if SEGMENT_ERROR:
                        status = "Error"

                    else:
                        status = "Unknown"
                segment_status_comment = status

        one_sim_status[sim_key].update({"Status": status})

        one_sim_status[sim_key].update({"Comment": segment_status_comment})

        one_sim_status[sim_key].update({"Allocation": allocation})

        return one_sim_status

    def construct_simulation_status(self):
        """Construct the simulation status of all the sims found

        Check if Ringdown segment
            Check TerminationCondition
                If finished
                If not CheckRunning
        Elseif Check Runnung
        Else NotRunning

        NotRunning:
            CheckFailed
        """

        sim_status = {}

        for sim_name in self.available_sim_names:
            # one_sim_status = {}
            for lev in self.available_sim_levs[sim_name]:
                one_status = self.read_last_segment_status(sim_name, lev)
                print(one_status)
                sim_status.update(one_status)

            # sim_status.update({sim_name: one_sim_status})

        self._sim_status = pd.DataFrame(sim_status).transpose().sort_index()

    def compute_ncycles(self):
        """Compute the number of cycles using the orbital
        revolutions"""

        import h5py

        all_sim_ncycles = {}

        for sim_name in self.available_sim_names:
            one_sim_ncycles = {}

            for lev in self.available_sim_levs[sim_name]:

                message(
                    f"Computing ncycles for {sim_name} Lev{lev}",
                    message_verbosity=2,
                )

                path_to_joined_h5 = self.join_horizons_file(sim_name, lev)

                try:
                    hf = h5py.File(path_to_joined_h5)
                except Exception as excep:
                    message(excep, f"Skipping {sim_name} Lev{lev}")
                    continue

                CA = hf["AhA.dir"]["CoordCenterInertial.dat"][...]
                CB = hf["AhB.dir"]["CoordCenterInertial.dat"][...]

                tA = CA[:, 0]
                xA = CA[:, 1]
                yA = CA[:, 2]
                zA = CA[:, 3]

                tB = CB[:, 0]
                xB = CB[:, 1]
                yB = CB[:, 2]
                zB = CB[:, 3]

                dx = xA - xB
                dy = yA - yB
                dz = zA - zB

                # Assuming low precession of orbital plane

                dRxy = dx + 1j * dy

                phase = np.unwrap(np.angle(dRxy))

                if np.mean(np.diff(phase)) < 0:
                    phase = -phase

                assert (
                    min(phase) == phase[0]
                ), "Phase must be min at the beginning"
                assert max(phase) == phase[-1], "Phase must be max at the end"

                phase = phase - min(phase)

                ncycles = max(phase / (2 * np.pi))

                one_sim_ncycles.update({lev: ncycles * 2})

                hf.close()

            all_sim_ncycles.update({sim_name: one_sim_ncycles})

        self._ncycles = all_sim_ncycles

    def compute_ncycles_from_waveform(self):
        """Compute the ncycles from the 2,2 mode waveform"""

        import h5py

        all_wf_ncycles = {}

        for sim_name in self.available_sim_names:
            one_wf_ncycles = {}

            for lev in self.available_sim_levs[sim_name]:

                message(
                    f"Computing ncycles for {sim_name} Lev{lev}",
                    message_verbosity=2,
                )

                path_to_extrap_waveform_h5 = self.get_waveform_out_path(
                    sim_name, lev
                )

                try:
                    hf = h5py.File(path_to_extrap_waveform_h5)

                except Exception as excep:
                    message(excep, f"Skipping {sim_name} Lev{lev}")
                    continue

                h22 = hf["Y_l2_m2.dat"][...]

                h22_re = h22[:, 1]
                h22_im = h22[:, 2]

                h22_cmplx = h22_re + 1j * h22_im

                h22_maxloc = np.argmax(abs(h22_cmplx))

                phase_22 = np.unwrap(np.angle(h22_cmplx[:h22_maxloc]))

                if np.mean(np.diff(phase_22)) < 0:
                    phase_22 = -phase_22

                phase_22 = phase_22 - min(phase_22)

                ncycles = max(phase_22 / (2 * np.pi))

                one_wf_ncycles.update({lev: ncycles})

                hf.close()

            all_wf_ncycles.update({sim_name: one_wf_ncycles})

        self._ncycles_wf = all_wf_ncycles

    def get_waveform_out_path(self, sim_name, lev, N=2):
        """Fetch the waveform output path"""

        wpath = (
            self.prepared_waveforms_dir / f"{sim_name}_waveforms_Lev{lev}/"
            f"extrapolated/rhOverM_Extrapolated_N{N}_CoM.h5"
        )

        return wpath

    def join_horizons_file(self, sim_name, lev):
        """Join the horizons.h5 file across segments for
        a simulation and return the path"""

        path_to_joined_h5 = (
            self.prepared_waveforms_dir
            / f"{sim_name}_waveforms_Lev{lev}/joined/"
            f"{sim_name}Lev{lev}JoinedHorizons.h5"
        )

        if os.path.exists(path_to_joined_h5):
            message(
                f"Joined Horizons.h5 file "
                "already found in {path_to_joined_h5}",
                message_verbosity=2,
            )

        else:
            wfp = PrepareSXSWaveform(
                sim_name=sim_name,
                sim_dir=self.available_sim_paths[sim_name],
                lev=lev,
                out_dir=self.prepared_waveforms_dir,
            )

            wfp.join_horizons()

            path_to_joined_h5 = wfp.joined_horizons_outfile_path

        return path_to_joined_h5

    def discover_ref_time(self):
        """Discover ref time as the light travel time to the
        innermost waveform extraction sphere"""

        all_req_ref_times = {}

        for sim_name, sim_path in self.available_sim_paths.items():

            try:
                one_sim_lev = self.available_sim_levs[sim_name][0]
            except IndexError:
                continue

            one_sim_ecc = self.get_ecc_dirs(sim_path)[0]

            sub_path = f"{one_sim_ecc}/Ev/Lev{one_sim_lev}_AA/Run/GW2/"
            "rh_FiniteRadii_CodeUnits.h5"

            wf_file_path = sim_path.joinpath(sub_path)

            with h5py.File(wf_file_path) as wf:

                print(sim_name)

                print(wf.keys())

                rkeys = [item for item in wf.keys() if ".dir" in item]

                ref_time = 2 * min([int(item[1:-4]) for item in rkeys])

                print(ref_time)

                all_req_ref_times.update({sim_name: ref_time})
                self._all_sim_params[sim_name].update(
                    {"RequestedReferenceTime": ref_time}
                )

        self._all_req_ref_times = all_req_ref_times

    def get_segment_at_req_ref_time(self):
        """Get the segments corresponding
        to the requested reference time"""

        all_req_ref_time_segments = {}

        for sim_name, sim_path in self.available_sim_paths.items():

            message(f"Looking in sim {sim_name}", message_verbosity=2)
            # try:

            sim_lev = 3
            if sim_lev not in self.available_sim_levs[sim_name]:
                sim_lev = min(self.available_sim_levs[sim_name])

            # highest_sim_lev = max(self.available_sim_levs[sim_name])
            highest_sim_ecc = max(self.get_ecc_dirs(sim_path))
            # except IndexError:
            #    continue

            seg_count = 0

            for seg in self.inspiral_segments[sim_name][f"Lev{sim_lev}"]:

                message(
                    f"Searching in Lev{sim_lev}_{seg}", message_verbosity=2
                )

                path_to_traj = f"{highest_sim_ecc}/Ev/Lev{sim_lev}_{seg}"
                "/Run/ApparentHorizons/Trajectory_AhA.dat"

                try:
                    traj_dat = np.genfromtxt(
                        self.available_sim_paths[sim_name].joinpath(
                            path_to_traj
                        )
                    )
                    avail_seg_times = traj_dat[:, 0]
                except (FileNotFoundError, IndexError) as fe:
                    message(f"Skipping segment {seg} as Traj file is missing")
                    seg_count += 1
                    continue

                # print(avail_seg_times)

                if (
                    self.all_req_ref_times[sim_name] < max(avail_seg_times)
                ) and (
                    self.all_req_ref_times[sim_name] > min(avail_seg_times)
                ):
                    prev_seg = self.inspiral_segments[sim_name][
                        f"Lev{sim_lev}"
                    ][seg_count]
                    all_req_ref_time_segments.update(
                        {sim_name: f"Lev{sim_lev}_{prev_seg}"}
                    )
                    message(
                        f"Segment found for {sim_name}", message_verbosity=2
                    )
                    break

                seg_count += 1

        self._all_req_ref_time_segments = all_req_ref_time_segments

    def parse_sim_params_at_req_ref_times(self):
        """Parse the masses, spins and Orbital freq at requested ref times"""

        from waveformtools.differentiate import derivative

        for sim_name, sim_seg in self.all_req_ref_time_segments.items():

            sim_path = self.available_sim_paths[sim_name]

            # for sim_name, sim_path in self.available_sim_paths.items():

            highest_sim_lev = max(self.available_sim_levs[sim_name])
            highest_sim_ecc = max(self.get_ecc_dirs(sim_path))

            req_ref_seg = self.all_req_ref_time_segments[sim_name]
            req_tref = self.all_req_ref_times[sim_name]

            sub_path_to_h = f"{highest_sim_ecc}/Ev/{req_ref_seg}/Run/"
            "ApparentHorizons/Horizons.h5"

            full_path_to_h = self.available_sim_paths[sim_name].joinpath(
                sub_path_to_h
            )

            with h5py.File(full_path_to_h) as hhf:

                message(sim_name)

                # print(hhf.keys())

                aha_dat = hhf["AhA.dir"]
                ahb_dat = hhf["AhB.dir"]

                irr_mass1_arr = aha_dat["ArealMass.dat"][...]
                chr_mass1_arr = aha_dat["ChristodoulouMass.dat"][...]
                chr_mass1 = get_val_at_t_ref(
                    chr_mass1_arr[:, 0], chr_mass1_arr[:, 1], req_tref
                )

                irr_mass2_arr = ahb_dat["ArealMass.dat"][...]
                chr_mass2_arr = ahb_dat["ChristodoulouMass.dat"][...]
                chr_mass2 = get_val_at_t_ref(
                    chr_mass2_arr[:, 0], chr_mass2_arr[:, 1], req_tref
                )

                spin_1_arr = aha_dat["DimensionfulInertialSpin.dat"][...]
                spin_2_arr = ahb_dat["DimensionfulInertialSpin.dat"][...]

                j_1x = get_val_at_t_ref(
                    spin_1_arr[:, 0], spin_1_arr[:, 1], req_tref
                )
                j_1y = get_val_at_t_ref(
                    spin_1_arr[:, 0], spin_1_arr[:, 2], req_tref
                )
                j_1z = get_val_at_t_ref(
                    spin_1_arr[:, 0], spin_1_arr[:, 3], req_tref
                )

                j_2x = get_val_at_t_ref(
                    spin_2_arr[:, 0], spin_2_arr[:, 1], req_tref
                )
                j_2y = get_val_at_t_ref(
                    spin_2_arr[:, 0], spin_2_arr[:, 2], req_tref
                )
                j_2z = get_val_at_t_ref(
                    spin_2_arr[:, 0], spin_2_arr[:, 3], req_tref
                )

                chi_1_arr = aha_dat["chiInertial.dat"][...]
                chi_1x = get_val_at_t_ref(
                    chi_1_arr[:, 0], chi_1_arr[:, 1], req_tref
                )
                chi_1y = get_val_at_t_ref(
                    chi_1_arr[:, 0], chi_1_arr[:, 2], req_tref
                )
                chi_1z = get_val_at_t_ref(
                    chi_1_arr[:, 0], chi_1_arr[:, 3], req_tref
                )

                chi_2_arr = ahb_dat["chiInertial.dat"][...]
                chi_2x = get_val_at_t_ref(
                    chi_2_arr[:, 0], chi_2_arr[:, 1], req_tref
                )
                chi_2y = get_val_at_t_ref(
                    chi_2_arr[:, 0], chi_2_arr[:, 2], req_tref
                )
                chi_2z = get_val_at_t_ref(
                    chi_2_arr[:, 0], chi_2_arr[:, 3], req_tref
                )

                # print('Spin1 shape', spin_1_arr.shape)
                # print('Chi1 shape', chi_1_arr.shape)

                # print("Diff1", irr_mass1_arr[:, 1] - chr_mass1_arr[:, 1])
                # print("Diff2", irr_mass2_arr[:, 1] - chr_mass2_arr[:, 1])

                subpath_to_traj1 = Path(
                    f"{highest_sim_ecc}/Ev/{req_ref_seg}/Run/ApparentHorizons/"
                    "Trajectory_AhA.dat"
                )
                subpath_to_traj2 = Path(
                    f"{highest_sim_ecc}/Ev/{req_ref_seg}/Run/ApparentHorizons/"
                    "Trajectory_AhB.dat"
                )

                full_path_to_traj1 = self.available_sim_paths[
                    sim_name
                ].joinpath(subpath_to_traj1)
                full_path_to_traj2 = self.available_sim_paths[
                    sim_name
                ].joinpath(subpath_to_traj2)

                traj1 = np.genfromtxt(full_path_to_traj1)
                traj2 = np.genfromtxt(full_path_to_traj2)

                x1 = traj1[:, 1]
                y1 = traj1[:, 2]

                x2 = traj2[:, 1]
                y2 = traj2[:, 2]

                r = (x2 - x1) + 1j * (y2 - y1)

                orb_phase = np.unwrap(np.angle(r))

                np.testing.assert_array_almost_equal(
                    traj1[:, 0],
                    traj2[:, 0],
                    12,
                    "Time axis from trajectory A and B must agree!",
                )

                orb_freq = derivative(traj1[:, 0], orb_phase, degree=5)

                omega_init = get_val_at_t_ref(traj1[:, 0], orb_freq, req_tref)

                chi1_mag = np.sqrt(chi_1x**2 + chi_1y**2 + chi_1z**2)
                chi2_mag = np.sqrt(chi_2x**2 + chi_2y**2 + chi_2z**2)

                j1_mag = np.sqrt(j_1x**2 + j_1y**2 + j_1z**2)
                j2_mag = np.sqrt(j_2x**2 + j_2y**2 + j_2z**2)

                hmass1 = np.sqrt(chr_mass1**2 + j1_mag**2 / (4 * chr_mass1**2))
                hmass2 = np.sqrt(chr_mass2**2 + j2_mag**2 / (4 * chr_mass2**2))

                message(f"Reference parameters at tref= {req_tref} are")
                message("------------------------")
                message(f" Christodlou Masses : M1 {chr_mass1} M2 {chr_mass2}")
                message(f" Horizon Masses : M1 {hmass1} M2 {hmass2}")
                message(
                    f" Chi magnitudes : |Chi1| {chi1_mag} |Chi2| {chi2_mag}"
                )
                message(f" J magnitudes : |J1| {j1_mag} |J2| {j2_mag}")
                message(
                    f" Chi : Chi1 ({chi_1x}, {chi_1y}, {chi_1z})"
                    f"Chi2 ({chi_2x}, {chi_2y}, {chi_2z})"
                )
                message(
                    f" J : J1 ({j_1x}, {j_1y}, {j_1z})"
                    f"Chi2 ({j_2x}, {j_2y}, {j_2z})"
                )
                message(f" Omega_i : {omega_init}")
                message("------------------------")

                import matplotlib.pyplot as plt

                # plt.plot(traj1[:, 0], orb_freq)
                # plt.show()

                # Update params

                self.all_sim_params[sim_name].update(
                    {"ReqRefTime": self.all_req_ref_times[sim_name]}
                )
                self.all_sim_params[sim_name].update({"Omega_ref": omega_init})
                self.all_sim_params[sim_name].update(
                    {"ChiA_ref": (chi_1x, chi_1y, chi_1z)}
                )
                self.all_sim_params[sim_name].update(
                    {"ChiB_ref": (chi_2x, chi_2y, chi_2z)}
                )
                self.all_sim_params[sim_name].update({"MA_ref": hmass1})
                self.all_sim_params[sim_name].update({"MB_ref": hmass2})

    def delete_empty_sims(self):
        """Delete sims that have no levels in them"""

        to_delete = []

        for sim_name, avail_levs in self.available_sim_levs.items():

            if avail_levs == []:

                to_delete.append(sim_name)

                try:
                    self._available_sim_names.remove(sim_name)
                    self._available_sim_paths.pop(sim_name)
                    # self._available_sim_levs.pop(sim_name)
                    self._all_sim_params.pop(sim_name)
                    self._all_sim_params_df.pop(sim_name)

                except Exception as ex:
                    message(ex, "Data not loaded")

        for item in to_delete:
            self._available_sim_levs.pop(item)

        self._deleted_sims = to_delete

    def run(self):
        """Run the pipeline"""
        self.search_simulations()
        self.discover_segments()
        self.discover_levels()
        self.delete_empty_sims()
        self.construct_simulation_status()

        self.fetch_sim_params()

        self.discover_ref_time()
        self.get_segment_at_req_ref_time()
        self.parse_sim_params_at_req_ref_times()

        self.update_pandas_dataframe()

    def get_filtered_dataframe(self):
        """Get only a subset of the data"""

        return self.all_sim_params_df.filter(items=self.filter_items)


def re_fetch_float(line):
    """Fetch a float value from .input file"""
    result = re.search("= -?[0-9]*[.]?[0-9]*", line)

    val = float(result.group()[2:])

    return val


def re_fetch_string(line):
    """Fetch a string value from .input file"""
    result = re.search('".*";', line)

    val = str(result.group()[1:-2])

    return val


def re_fetch_vector(line):
    """Fetch a vector value from a .input file"""
    # result = re.search('\(-?[0-9]*[.]?[0-9]*,
    # -?[0-9]*[.]?[0-9]*,-?[0-9]*[.]?[0-9]*\)' ,line)
    result1 = re.search("\(-?[0-9]*[.]?[0-9]*,", line)
    result2 = re.search(",-?[0-9]*[.]?[0-9]*,", line)
    result3 = re.search(",-?[0-9]*[.]?[0-9]*\)", line)

    # val = str(result.group())
    val1 = float(result1.group()[1:-1])
    val2 = float(result2.group()[1:-1])
    val3 = float(result3.group()[1:-1])

    return (val1, val2, val3)


def re_fetch_ext_rad(line):
    """Fetch the ext radius from the key"""


def get_val_at_t_ref(time_axis, val_axis, time):
    """Interpolate and get the value at the requested time"""

    from scipy.interpolate import interp1d

    int_func = interp1d(time_axis, val_axis, kind="cubic")

    val_at_t_ref = float(int_func(time))

    # print(val_at_t_ref, type(val_at_t_ref))

    return round(val_at_t_ref, 5)



""" Diagnostic tools """

import numpy as np


class method_info:
    """The methods for integration ,differential to be passed on
    for operations."""

    def __init__(
        self,
        int_method="MP",
        diff_method="SH",
        ell_max=8,
        eval_ell_max=8,
        degree=8,
        reg=True,
        swsh_routine="waveformtools",
        atol=1e-4,
        rtol=1e-2,
        ode_solver="solve_ivp",
        ivp_method="RK45",
        vectorized=True,
        error_info=False,
    ):

        self.int_method = int_method
        self.diff_method = diff_method
        self.ell_max = ell_max
        self.eval_ell_max = eval_ell_max
        self.reg = reg
        self.degree = degree
        self.swsh_routine = swsh_routine
        self.atol = atol
        self.rtol = rtol
        self.ode_solver = ode_solver
        self.vectorized = vectorized
        self.ivp_method = ivp_method
        self.error_info = error_info


def RMSerrs(func1, func2, info):
    """Compute and return the RMS error between two arrays

    Parameters
    ----------
    func1, func2 : ndarray
                   Arrays of same shape to compare with.
    info : sphericalarray
           Grid info

    Returns
    -------
    RMS : float
          The RMS error
    Amax : float
           The max diff
    Amin : float the min diff
    """
    diff = func1 - func2

    Amax = np.amax(diff)
    Amin = np.amin(diff)

    RMS = np.sqrt(
        np.sum(np.absolute(diff) ** 2, axis=(-2, -1)) / info.npix_act
    )

    return RMS, Amin, Amax


def IsModesEqual(modes1, modes2, modes_list):
    """Check if the mode amplitudes are equal

    Parameters
    ----------
    modes1, modes2: dict
                    A dictionary of modes.
                    For each mode, the data could be an array.

    modes_list: list
                A list of list of mode numbers
    Returns
    -------
    modes_err: float
               The error of each mode as a dict
    """

    modes_err = {}

    for ell, emm_list in modes_list:
        for emm in emm_list:
            A1 = modes1[f"l{ell}"][f"m{emm}"]
            A2 = modes2[f"l{ell}"][f"m{emm}"]

            delta = A1 - A2

            label = f"l{ell}m{emm}"

            modes_err.update({label: delta})

    return modes_err



""" Methods to integrate functions """

#################################################
# Imports
################################################
import numpy as np
import scipy
from spectral.fourier.fft import compute_ifft
from waveformtools.waveformtools import message

##################################################
# Fixed frequency integration
##################################################


def fixed_frequency_integrator(
    udata_time,
    delta_t,
    utilde_conven=None,
    freq_axis=None,
    omega0=0,
    order=1,
    zero_mode=0,
):
    """Fixed frequency integrator as presented in Reisswig et. al.

    Parameters
    ----------
    udata_time:	1d array
                The input data in time.
    delta_t: float
             The time stepping.

    utilde_conven: 1d array, optional
                   The conventional FFT of the samples udata_time.
    freq_axis: 1darray, optional
               The frequency axis of the FFT. Must be supplied
               along with `utilde_conven`.
    omega0:	float, optional
            The cutoff angular frequency in the integration.
            Must be lower than the starting angular frequency
            of the input waveform. All frequencies whose absolute
            value is below this value will be neglected.
            The default cutoff-value is 0.

    order: int, optional
           The number of times to integrate
           the integrand in time. Defaults to 1.

    zero_mode: float, optional
               The zero mode amplitude of the FFT required.
               Defaults to 0 i.e. the zero mode is removed.

    Returns
    -------
    u_integ_n_time:	1d array
                    The input waveform in time-space,
                    integrated in frequency space using FFI.

    u_integ_integ_n: 1d array
                     The integrated u samples in Fourier space.

    """

    if not utilde_conven:
        # Compute the FFT of data
        from numpy.fft import ifft
        from spectral.fourier.fft import compute_fft, unset_fft_conven

        # from waveformtools import taper
        # udata_x_re = taper(u_time.real, delta_t=delta_t)
        # udata_x_im = taper(u_time.imag, delta_t=delta_t)
        # udata_x	   = np.array(udata_x_re) + 1j * np.array(udata_x_im)
        # x_axis = udata_x_re.sample_times
        # udata_x = np.array(udata_x)
        freq_axis, utilde_conven = compute_fft(udata_time, delta_t)

        if np.isnan(utilde_conven).any():
            message("Nan Found in utilde_conven!")

        if np.isnan(freq_axis).any():
            message("Nan Found in freq_axis!")

        # Find the length of the input data.
        Nlen = len(udata_time)

    else:
        Nlen = len(utilde_conven)
        assert (
            np.array(freq_axis) != np.array(None)
        ).all(), "Please supply the frequency axis along with utilde_conven"

    # df = np.diff(freq_axis)[0]
    df = scipy.stats.mode(np.diff(freq_axis))[0]
    message("df ", df, message_verbosity=3)

    # Find the location of the zero index.
    if (freq_axis == 0).any():
        zero_index = np.where(freq_axis == 0)[0][0]
    else:
        zero_index = None

    omega_axis = construct_ffi_omega_axis(freq_axis, omega0, zero_index)

    # Set the zero frequency element separately.
    utilde_conven[zero_index] = zero_mode

    # Integrate in frequency space
    utilde_integ_n = np.power((-1j / omega_axis), order) * utilde_conven

    if np.isnan(utilde_integ_n).any():
        message("Nan Found in utilde_integ_n!", message_verbosity=0)

    # Get the inverse fft
    # utilde_integ_n_orig = unset_fft_conven(utilde_integ_n)

    # u_integ_n_time = ifft(utilde_integ_n_orig)
    u_integ_n_time, u_integ = compute_ifft(utilde_integ_n, df)

    return u_integ_n_time, u_integ


def construct_ffi_omega_axis(freq_axis, omega0, zero_index):
    """Construct an angular frequency axis for
    use with FFI"""

    assert (
        omega0 > 0
    ), "Please supply a non-zero positive value of cutoff angular frequency"

    # Construct the angular frequency axis.
    omega_axis = 2 * np.pi * freq_axis
    # omega_axis[zero_index] = 1

    message(
        "The chosen cutoff angular frequency is", omega0, message_verbosity=2
    )
    message("Omega axis", omega_axis, message_verbosity=4)

    # Change the angular frequency if its magnitude is below a given omega0.
    for index, element in enumerate(omega_axis):
        # Loop over the samples.
        # Skip the zero index
        if index != zero_index:
            # print(freq_integ[index])
            # try:
            # Get the sign of the angular frequency.
            sign = int(element / abs(element))
            if abs(element) < omega0:
                omega_axis[index] = sign * omega0

        else:
            sign = 1
            assert (
                omega_axis[index] == 0
            ), f"The zero mode element must be zero frequency. Instead it is {omega_axis[index]}"
            omega_axis[index] = 1

    return omega_axis


#############################################
# 2D integrals
#############################################


def TwoDIntegral(func, grid_info, int_method=None):
    """Integrate a function over a sphere.

    Parameters
    ----------
    func: function
          The function to be integrated
    NTheta, NPhi: int
                  The number of grid points in the theta and phi directions.
                  Note that NTheta must be even.
    ht, hp: float
            The grid spacings.
    int_method: string
                The method to use for the integration.
                Options are DH (Driscoll Healy),
                SP (Simpson's),
                MP (Midpoint).

    Returns
    -------
    integ : float
            The function f integrated over the sphere.
    """

    if int_method is None:
        if grid_info.grid_type == "GL":
            int_method = "GL"
        elif grid_info.grid_type == "Uniform":
            int_method = "MP"
        else:
            raise KeyError(
                "Unable to discern default integration"
                "method due to unknown grid type. Please provide"
                "method explicitely"
            )

    if int_method == "DH":
        integral = DriscollHealy2DInteg(func, grid_info)
    elif int_method == "MP":
        integral = MidPoint2DInteg(func, grid_info)
    elif int_method == "SP":
        integral = Simpson2DInteg(func, grid_info)
    elif int_method == "GL":
        integral = GaussLegendre2DInteg(func, grid_info)

    else:
        raise ValueError("Unknown method!")
    return integral


def MidPoint2DInteg(func, info):
    """Evaulate the 2D surface integral using the midpoint rule.

    Parameters
    ----------
    func: ndarray
          The data to be integrated
    info: surface_grid_info
          An instance of the surface grid info class
          containing information about the grid.
    Returns
    -------
    integ: float
           The function f integrated over the sphere.
    """

    ht = info.dtheta
    hp = info.dphi

    theta_grid, _ = info.meshgrid

    integral = (
        np.tensordot(func, np.sin(theta_grid), axes=((-2, -1), (0, 1)))
        * ht
        * hp
    )

    return integral


def DriscollHealy2DInteg(func, info):
    """Implementation of the Driscoll Healy 2D integration that
    exhibits near spectral convergence.

    Parameters
    ----------
    func : function
            The function to be integrated
    NTheta, NPhi : int
             The number of grid points in the theta and phi directions.
             Note that NTheta must be even.
    ht, hp : float
             The grid spacings.

    Returns
    -------
    integ : float
            The function f integrated over the sphere.
    """
    if len(func.shape) > 2:
        raise NotImplementedError(
            "Driscoll-Healy's method cannot currently handle 3d arrays"
        )

    NTheta = info.ntheta_act
    NPhi = info.nphi_act

    # NTheta, NPhi = func.shape
    theta_grid, _ = info.meshgrid

    ht = info.dtheta
    hp = info.dphi

    if NTheta < 0:
        raise ValueError("Ntheta is negative!")
    elif NPhi < 0:
        raise ValueError("Nphi is negative!")

    if (NTheta % 2) != 0:
        raise ValueError("NTheta must be even!")

    integrand_sum = 0.0

    func *= np.sin(theta_grid)

    # Skip the poles (ix=0 and ix=NTheta), as the weight there is zero
    # theta_1d = np.pi* np.arange(1, NTheta)/NTheta
    # ell_weight_axis = np.arange(int(NTheta/2))
    # theta_2d, ell_2d = np.meshgrid(theta_1d, ell_weight_axis)
    # weights_grid = (4/ np.pi) * np.sin((2 * ell_2d + 1) * theta_2d) /
    # (2 * ell_2d + 1)
    # weights_axis = np.sum(weights_grid, axis=1)
    # latitude_sum_axis = np.sum(func, axis=1)
    # integrand_axis = latitude_sum_axis * weights_axis

    for theta_index in range(1, NTheta):
        # These weights lead to an almost spectral convergence
        this_theta = np.pi * theta_index / NTheta

        # this_theta = theta_1d[theta_index]

        weight = 0

        # theta = M_PI * ix / NTheta;
        # weight = 0.0;
        for ell in range(int(NTheta / 2)):
            # for (int l = 0; l < NTheta/2; ++ l)
            weight += np.sin((2 * ell + 1) * this_theta) / (2 * ell + 1)

        # weight_axis = np.sin((2 * ell_weight_axis + 1)
        # * this_theta) / (2 * ell_weight_axis + 1)

        weight *= 4.0 / np.pi
        latitude_sum = 0

        # local_sum = 0.0;
        # Skip the last point (iy=NPhi), since we assume periodicity and
        # therefore it has the same value as the first point. We don't use
        # weights in this direction, which leads to spectral convergence.
        # (Yay periodicity!)

        for index_phi in range(NPhi):
            # for (int iy = 0; iy < NPhi; ++ iy)
            latitude_sum += func[theta_index, index_phi]

        # latitude_sum = np.sum(func[index_theta, :])

        integrand_sum += weight * latitude_sum

    return ht * hp * integrand_sum


def DriscollHealy2DInteg_v2(func, info):
    """Implementation of the Driscoll Healy 2D integration that
    exhibits near spectral convergence.

    Parameters
    ----------
    func : function
            The function to be integrated
    NTheta, NPhi : int
             The number of grid points in the theta and phi directions.
             Note that NTheta must be even.
    ht, hp : float
             The grid spacings.

    Returns
    -------
    integ : float
            The function f integrated over the sphere.
    """

    if len(func.shape) > 2:
        raise NotImplementedError(
            "Driscoll-Healy's method cannot currently handle 3d arrays"
        )

    NTheta = info.ntheta_act
    NPhi = info.nphi_act

    # NTheta, NPhi = func.shape
    theta_grid, _ = info.meshgrid

    ht = info.dtheta
    hp = info.dphi

    if NTheta < 0:
        raise ValueError("Ntheta is negative!")
    elif NPhi < 0:
        raise ValueError("Nphi is negative!")

    if (NTheta % 2) != 0:
        raise ValueError("NTheta must be even!")

    integrand_sum = 0.0

    # func*=np.sin(theta_grid)

    # Skip the poles (ix=0 and ix=NTheta), as the weight there is zero

    theta_1d = np.pi * np.arange(1, NTheta) / NTheta
    ell_weight_axis = np.arange(int(NTheta / 2))

    print("theta 1d", len(theta_1d))
    print("ell axis", len(ell_weight_axis))

    theta_2d, ell_2d = np.meshgrid(theta_1d, ell_weight_axis)
    print("T2d", theta_2d.shape)

    weights_grid = (
        (4 / np.pi) * np.sin((2 * ell_2d + 1) * theta_2d) / (2 * ell_2d + 1)
    )

    print("WG", weights_grid.shape)

    weights_axis = np.sum(weights_grid, axis=1)
    print("WG", weights_axis.shape)

    latitude_sum_axis = np.sum(func, axis=1)
    print("LSA", latitude_sum_axis.shape)

    integrand_axis = latitude_sum_axis * weights_axis

    integrand_sum = np.sum(integrand_axis)

    return ht * hp * integrand_sum


def Simpson2DInteg(func, info):
    """Implementation of Simpson's 2D integration
    scheme.

    Parameters
    ----------
    func : function
            The function to be integrated
    NTheta, NPhi : int
             The number of grid points in the theta and phi directions.
             Note that NTheta must be even.
    ht, hp : float
             The grid spacings.

    Returns
    -------
    integ : float
            The function f integrated over the sphere.
    """

    if len(func.shape) > 2:
        raise NotImplementedError(
            "Simpson's method cannot currently handle 3d arrays"
        )

    NTheta = info.ntheta_act
    NPhi = info.nphi_act

    theta_grid, _ = info.meshgrid
    # NTheta, NPhi = func.shape

    ht = info.dtheta
    hp = info.dphi

    integrand_sum = 0
    index_theta = 0
    index_phi = 0

    assert NTheta > 0
    assert NPhi > 0
    # assert(func.all())
    assert NTheta % 2 == 0
    assert NPhi % 2 == 0

    Px = int(NTheta / 2)
    Py = int(NPhi / 2)

    func *= np.sin(theta_grid)

    # Around corners
    integrand_sum += (
        func[0, 0]
        + func[NTheta - 1, 0]
        + func[0, NPhi - 1]
        + func[NTheta - 1, NPhi - 1]
    )

    # Arount edges
    for index_phi in range(1, Py):
        integrand_sum += (
            4 * func[0, 2 * index_phi - 1]
            + 4 * func[NTheta - 1, 2 * index_phi - 1]
        )

    # for (iy = 1; iy <= py-1; iy++)
    for index_phi in range(1, Py - 1):
        integrand_sum += (
            2 * func[0, 2 * index_phi] + 2 * func[NTheta - 1, 2 * index_phi]
        )

    # for (ix = 1; ix <= px; ix++)
    for index_theta in range(1, Px):
        integrand_sum += (
            4 * func[2 * index_theta - 1, 0]
            + 4 * func[2 * index_theta - 1, NPhi - 1]
        )

    # for (ix = 1; ix <= px-1; ix++)
    for index_theta in range(1, Px - 1):
        integrand_sum += (
            2 * func[2 * index_theta, 0] + 2 * func[2 * index_theta, NPhi - 1]
        )

    # In the Interiors
    # for (iy = 1; iy <= py; iy++)
    for index_phi in range(1, Py):
        # for (ix = 1; ix <= px; ix++)
        for index_theta in range(1, Px):
            integrand_sum += 16 * func[2 * index_theta - 1, 2 * index_phi - 1]

    # for (iy = 1; iy <= py-1; iy++)
    for index_phi in range(1, Py - 1):
        # for (ix = 1; ix <= px; ix++)
        for index_theta in range(Px):
            integrand_sum += 8 * func[2 * index_theta - 1, 2 * index_phi]

    # for (iy = 1; iy <= py; iy++)
    for index_phi in range(1, Py):
        # for (ix = 1; ix <= px-1; ix++)
        for index_theta in range(1, Px - 1):
            integrand_sum += 8 * func[2 * index_theta, 2 * index_phi - 1]

    # for (iy = 1; iy <= py-1; iy++)
    for index_phi in range(1, Py - 1):
        # for (ix = 1; ix <= px-1; ix++)
        for index_theta in range(1, Px - 1):
            integrand_sum += 4 * func[2 * index_theta, 2 * index_phi]

    return (1 / 9) * ht * hp * integrand_sum


def GaussLegendre2DInteg(func, info):
    """Evaulate the 2D surface integral using the Gauss-Legendre rule.

    Parameters
    ----------
    func: ndarray
          The data to be integrated
    info: surface_grid_info
           An instance of the surface grid info class
           containing information about the grid.
    Returns
    -------
    integ: float
           The function f integrated over the sphere.
    """

    integral = (
        np.tensordot(func, info.weights_grid, axes=((-2, -1), (0, 1)))
        * info.dphi
    )
    return integral



import numpy as np
import os
from pathlib import Path
import config
import subprocess

import matplotlib.pyplot as plt

config.conf_matplolib()


from waveformtools.sxs.discover_sims import SimulationExplorer

list_of_search_dirs = ["/mnt/pfs/vaishak.p/sims/SpEC/gcc/bfi/ICTSEccParallel"]

se = SimulationExplorer(
    search_dir=list_of_search_dirs[0],
    prepared_waveforms_dir=Path("/mnt/pfs/vaishak.p/Projects/Codes/waveforms"),
)

se.search_simulations()
se.find_highest_ecc_dirs()
se.discover_levels()

# se.prepare_waveforms()


for sim_name, sim_path in se.found_sim_paths.items():

    inside_ev = sim_path.joinpath(f"Ecc{se.highest_ecc_nums[sim_name]}/Ev")

    out_file = sim_path.joinpath("WaveformsJoinedLev3.h5")

    cmd = f"JoinH5 -v -o {out_file} -l {inside_ev}/Lev3*/Run/GW2/rh_FiniteRadii_CodeUnits.h5"

    print(cmd)

    (cmd)



""" 
MIT License

Copyright (c) 2023 Vaishak Prasad

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


Script to 

    a) join h5 files
    b) extrapolate to infinity
    c) correct for CoM drift
    
    from SXS simulations
    
"""

import os
from pathlib import Path

import numpy as np
import scri
from waveformtools.waveformtools import message


class PrepareSXSWaveform:
    """Prepare waveforms from a particular Lev and ECC of
    an SXS NR run by

    1) Joining all segments
    2) Extrapolating to infinity
    3) Transforming to CoM frame
    4) Upload to cloud


    Attributes
    ----------
    sim_dir: str/POSIXPath
              The root directory containing
              the simulation directory
    sim_name: str
              The alias of the simulation whose
              waveform is to be processed
    out_dir: str/Path
             The directory in which to save extrapolated
             waveforms
    joined_outfile_dir: str/Path
                        The directory in which to save joined
                        waveforms from across segments
    joined_waveform_outfile_name: str/Path
                                  The filename of the processed
                                  waveform.
    joined_waveform_outfile_path: The full path to the processed
                                  waveform file.
    joined_horizons_outfile_name: str/Path
                                  The filename of the joined horizon file
    lev: int
         The sim resolution level to process
    ecc: int
         The Ecc dir number to process
    extrap_out_dir: str/Path
                    The full path to the output dir.

    Methods
    -------
    join_waveform_h5_files
        Join the waveform h5 files from across segments
    extrapolate
        Extrapolate a joined waveform to infinity
    join_horizons
        Join the Horizons.h5 files from across segments
    transform_to_CoM_frame
        Apply CoM correction
    upload_output_dir
        Upload the output directory to a cloud
    prepare_waveform
        Process the waveform by extrapolating and
        applying CoM correction
    """

    def __init__(
        self,
        sim_name,
        history_file=Path("./history_file.txt"),
        sim_dir=Path("./"),
        out_dir=None,
        joined_waveform_outfile_name=None,
        joined_horizons_outfile_name=None,
        lev=2,
        ecc=0,
    ):
        if not os.path.isabs(sim_dir):
            raise ValueError("Please provide the full sim path!")

        self._sim_dir = Path(sim_dir)

        self._sim_name = sim_name
        self._lev = lev
        self._ecc = ecc
        self._history_file = history_file

        if joined_waveform_outfile_name is None:
            message("Choosing default directory for output...")
            joined_waveform_outfile_name = (
                sim_name + f"Lev{self.lev}JoinedWaveform.h5"
            )

        self._joined_waveform_outfile_name = joined_waveform_outfile_name

        if out_dir is None:

            cwd = Path(os.getcwd())
            message(f"Current working directory {cwd}")
            self._out_dir = (
                cwd / f"processed_waveforms/{sim_name}_waveforms_Lev{self.lev}"
            )
            message(f"Setting out-directory to ({self.out_dir})...")
        else:
            self._out_dir = out_dir

        message(f"Out directory is set to {self.out_dir}")

        if not os.path.isdir(self.out_dir):
            message(f"Creating directory {self.out_dir}")
            os.makedirs(self.out_dir)

        joined_outfile_dir = os.path.join(self.out_dir, Path(f"joined"))

        self._joined_outfile_dir = joined_outfile_dir

        if not os.path.isdir(joined_outfile_dir):
            os.mkdir(joined_outfile_dir)

        self._joined_outfile_path = os.path.join(
            joined_outfile_dir, self.joined_waveform_outfile_name
        )

        if joined_horizons_outfile_name is None:
            self._joined_horizons_outfile_name = (
                sim_name + f"Lev{self.lev}JoinedHorizons.h5"
            )

        self.setup_env()

    @property
    def sim_dir(self):
        """The full path to the directory containing the
        simulation"""
        return self._sim_dir

    @property
    def sim_name(self):
        return self._sim_name

    @property
    def out_dir(self):
        return self._out_dir

    @property
    def joined_outfile_dir(self):
        """The directory containing the
        joined files"""
        return self._joined_outfile_dir

    @property
    def joined_waveform_outfile_name(self):
        return self._joined_waveform_outfile_name

    @property
    def joined_waveform_outfile_path(self):
        return os.path.join(
            self.joined_outfile_dir,
            Path(f"{self.joined_waveform_outfile_name}"),
        )

    @property
    def joined_horizons_outfile_name(self):
        return self._joined_horizons_outfile_name

    @property
    def joined_horizons_outfile_path(self):
        return os.path.join(
            self.joined_outfile_dir,
            Path(f"{self.joined_horizons_outfile_name}"),
        )

    @property
    def lev(self):
        return self._lev

    @property
    def ecc(self):
        return self._ecc

    @property
    def extrap_out_dir(self):
        return os.path.join(self.out_dir, Path("extrapolated"))

    @property
    def history_file(self):
        return self._history_file

    def setup_env(self):
        """Setup the environment"""

        # exec(open('/mnt/pfs/vaishak.p/soft/modules-5.2.0/init/python.py').read())

        # module('unload', 'gcc/11.1.0')

        # module('load', 'gcc/11.1.0')

    def join_waveform_h5_files(self, verbose=False):
        """Join the waveform h5 files"""

        if Path(self.joined_horizons_outfile_path).exists():
            message("File already exists. Skipping operation.")

        else:
            message("Joining waveform h5 files...")

            data_paths_insp = os.path.join(
                self.sim_dir,
                Path(
                    f"Ecc{self.ecc}"
                    f"/Ev/Lev{self.lev}*/Run/GW2/"
                    "rh_FiniteRadii_CodeUnits.h5"
                ),
            )

            data_paths_rdown = os.path.join(
                self.sim_dir,
                Path(
                    f"Ecc{self.ecc}"
                    f"/Ev/Lev{self.lev}_Ringdown/"
                    f"Lev{self.lev}*/Run/GW2/"
                    "rh_FiniteRadii_CodeUnits.h5"
                ),
            )

            if verbose:
                run_cmd = "JoinH5 -v"

            else:
                run_cmd = "JoinH5"

            try:
                run_cmd += (
                    f" -o {self.joined_waveform_outfile_path}"
                    f" {data_paths_insp} "
                    f" {data_paths_rdown}"
                )

                message(f"Running command\n {run_cmd}")

                # with open('join_waveforms_output.txt', "wb") as fout:
                # subprocess.check_call('dir',stdout=f)
                # cmd_out = subprocess.Popen(run_cmd, stdout=subprocess.PIPE)

                # for cline in iter(lambda: process.stdout.read(1), b""):
                # sys.stdout.buffer.write(cline)
                # f.buffer.write(cline)

                cmd = os.popen(run_cmd)

                out = cmd.read()

                message("Command output \n", out)
            except Exception as ex:
                run_cmd += (
                    f" -o {self.joined_waveform_outfile_path}"
                    f" {data_paths_insp}"
                )

                message(f"Running command\n {run_cmd}")

                # with open('join_waveforms_output.txt', "wb") as fout:
                # subprocess.check_call('dir',stdout=f)
                # cmd_out = subprocess.Popen(run_cmd, stdout=subprocess.PIPE)

                # for cline in iter(lambda: process.stdout.read(1), b""):
                # sys.stdout.buffer.write(cline)
                # f.buffer.write(cline)

                cmd = os.popen(run_cmd)

                out = cmd.read()

                message("Command output \n", out)

            message("Command completed. Please check Errors.txt for details")

    def extrapolate(self, ChMass=1.0, UseStupidNRARFormat=True):
        """Extrapolate the waveform"""

        try:
            message(
                "\tChecking if extrapolated files from a previous run exists..."
            )
            files = os.listdir(self.extrap_out_dir)

            exists = np.array(
                [item for item in files if "Extrapolated" in item]
            )
        # message(exists)
        except Exception as excep:
            message("\t", excep)
            message("\tNo extrapolated files from previous run found")
            exists = []

        if len(exists) > 0:
            message("\tSkipping extrapolation")

        else:
            message("Extrapolating...")

            wf = scri.extrapolate(
                InputDirectory=self.joined_outfile_dir,
                OutputDirectory=self.extrap_out_dir,
                DataFile=self.joined_waveform_outfile_name,
                ChMass=ChMass,
                UseStupidNRARFormat=UseStupidNRARFormat,
                DifferenceFiles="",
                PlotFormat="",
            )

    def join_horizons(self, verbose=False):
        """Join horizons file and save to the joined
        file dir"""

        if Path(self.joined_horizons_outfile_path).exists():
            message("File already exists. Skipping join horizons operation.")

        else:
            message("Joining Horizon h5 files...")

            input_insp_dat_rel_loc = (
                f"Ecc{self.ecc}"
                f"/Ev/Lev{self.lev}*/Run/"
                "ApparentHorizons/Horizons.h5"
            )

            input_rdown_dat_rel_loc = (
                f"Ecc{self.ecc}"
                f"/Ev/Lev{self.lev}_Ringdown/"
                f"Lev{self.lev}*/Run/"
                "ApparentHorizons/Horizons.h5"
            )

            # data_paths_insp = os.path.join(self.sim_dir, input_insp_dat_rel_loc)

            data_paths_insp = self.sim_dir.joinpath(input_insp_dat_rel_loc)

            message("data paths insp", data_paths_insp, message_verbosity=2)

            # data_paths_rdown = os.path.join(
            #    self.sim_dir, input_rdown_dat_rel_loc
            # )

            data_paths_rdown = self.sim_dir.joinpath(input_rdown_dat_rel_loc)

            message("data paths rdown", data_paths_rdown, message_verbosity=2)

            if verbose:
                run_cmd = "JoinH5 -v"

            else:
                run_cmd = "JoinH5"

            try:
                run_cmd += (
                    f" -o {self.joined_horizons_outfile_path}"
                    f" {data_paths_insp} {data_paths_rdown}"
                )

                message(f"Running command\n {run_cmd}")

                cmd = os.popen(run_cmd)

                out = cmd.read()

                message("Command output \n", out)

            except Exception as ex:
                run_cmd += (
                    f" -o {self.joined_horizons_outfile_path}"
                    f" {data_paths_insp}"
                )

                message(f"Running command\n {run_cmd}")

                cmd = os.popen(run_cmd)

                out = cmd.read()

                message("Command output \n", out)

            message("Command completed. Please check Errors.txt for details.")

    def transform_to_CoM_frame(
        self,
        skip_beginning_fraction=0.01,
        skip_ending_fraction=0.10,
        file_format="NRAR",
        extrap_enn_list=[-1, 2, 3, 4, 5, 6],
    ):
        """Apply CoM correction to a waveform"""
        from scri.SpEC.com_motion import remove_avg_com_motion

        try:
            files = os.listdir(self.extrap_out_dir)

            exists = np.array([item for item in files if "CoM" in item])

        except Exception as excep:
            message(excep)
            message("Continuing with transformation")
            exists = []

        if len(exists) > 0:
            message("Skipping CoM transformation")
        else:
            message("Transforming to CoM frame...")

            for extrap_enn in extrap_enn_list:
                message(f"Working on Extrapolated N_{extrap_enn}")

                path_to_waveform_h5 = os.path.join(
                    self.extrap_out_dir,
                    f"rhOverM_Extrapolated_N{extrap_enn}.h5",
                )

                path_to_horizons_h5 = self.joined_horizons_outfile_path

                remove_avg_com_motion(
                    w_m=None,
                    path_to_waveform_h5=path_to_waveform_h5,
                    path_to_horizons_h5=path_to_horizons_h5,
                    skip_beginning_fraction=skip_beginning_fraction,
                    skip_ending_fraction=skip_ending_fraction,
                    file_write_mode="w",
                    m_A=None,
                    m_B=None,
                    file_format=file_format,
                    write_corrected_file=True,
                )

    def upload_output_dir(self):
        """Upload the outut directory to a cloud"""
        raise NotImplementedError

    def prepare_waveform(
        self,
        verbose=False,
        ChMass=1.0,
        UseStupidNRARFormat=True,
        skip_beginning_fraction=0.01,
        skip_ending_fraction=0.10,
        file_format="NRAR",
        extrap_enn_list=[-1, 2, 3, 4, 5, 6],
        upload=False,
    ):
        """Carry out extrapolation + CoM correction"""
        self.join_waveform_h5_files(verbose=verbose)

        self.extrapolate(
            ChMass=ChMass, UseStupidNRARFormat=UseStupidNRARFormat
        )

        self.join_horizons(verbose=verbose)

        self.transform_to_CoM_frame(
            skip_beginning_fraction=skip_beginning_fraction,
            skip_ending_fraction=skip_ending_fraction,
            file_format=file_format,
            extrap_enn_list=extrap_enn_list,
        )

        if upload:
            self.upload_output_dir()

            pass

        message("\n--------------------------------------------------------\n")

        return True

    def write_history(self):
        """Write processing history to a file"""
        with open(self.history_file, "a") as th:
            for item in dirs:
                th.write(item)
                th.write("\n")

        # th.write('EOF')



import numpy as np
import unittest

from waveformtools.transforms import Yslm_vec
from waveformtools.grids import GLGrid

info = GLGrid()
theta_grid, phi_grid = info.meshgrid


fun1_ana_110 = np.sqrt(3 / (8 * np.pi)) * np.sin(theta_grid)
fun1_ana_11p1 = (
    -np.sqrt(3 / (16 * np.pi))
    * (1 - np.cos(theta_grid))
    * np.exp(1j * phi_grid)
)
fun1_ana_11m1 = (
    -np.sqrt(3 / (16 * np.pi))
    * (1 + np.cos(theta_grid))
    * np.exp(-1j * phi_grid)
)


class TestYslm(unittest.TestCase):
    def test_s1l1m0(self):
        """Test :math:`Y_{slm}` for :math:`s=1, \\ell=1, m=0`"""

        spin_weight = 1

        ell = 1

        emm1 = 0

        fun1_110 = Yslm_vec(
            spin_weight=spin_weight,
            ell=ell,
            emm=emm1,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )

        np.testing.assert_almost_equal(
            fun1_110,
            fun1_ana_110,
            15,
            "Yslm must agree upto 15 decimals for" "s=1, l=1, m=0",
        )

    def test_s1l1mp1(self):
        """Test :math:`Y_{slm}` for :math:`s=1, \\ell=1, m=1`"""

        spin_weight = 1

        ell = 1

        emm2 = 1

        fun1_11p1 = Yslm_vec(
            spin_weight=spin_weight,
            ell=ell,
            emm=emm2,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )

        np.testing.assert_almost_equal(
            fun1_11p1,
            fun1_ana_11p1,
            15,
            "Yslm must agree upto 15 decimals for" "s=1, l=1, m=1",
        )

    def test_s1l1mm1(self):
        """Test :math:`Y_{slm}` for :math:`s=1, \\ell=1, m=-1`"""

        spin_weight = 1

        ell = 1

        emm3 = -1

        fun1_11m1 = Yslm_vec(
            spin_weight=spin_weight,
            ell=ell,
            emm=emm3,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )

        np.testing.assert_almost_equal(
            fun1_11m1,
            fun1_ana_11m1,
            15,
            "Yslm must agree upto 15 decimals for" "s=1, l=1, m=-1",
        )


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False, verbosity=2)



import numpy as np
from spectral.spherical.transforms import SHExpand
from spectral.spherical.grids import GLGrid
from waveformtools.diagnostics import method_info
from waveformtools.waveformtools import message
import unittest, pytest
from config.verbosity import levels

lv = levels()
lv.set_print_verbosity(2)


class TestGLGridYlm(unittest.TestCase):
    def test_ylm_single_mode_recovery(self):
        """Test single mode recovery for every mode
        upto a given L"""

        info = GLGrid(L=11)
        minfo = method_info(ell_max=11, int_method="GL")

        from spectral.spherical.swsh import Yslm_vec

        theta_grid, phi_grid = info.meshgrid

        for ell in range(info.L + 1):
            for emm in range(-ell, ell + 1):
                message("Testing", ell, emm, message_verbosity=2)

                Ylm = Yslm_vec(
                    spin_weight=0,
                    ell=ell,
                    emm=emm,
                    theta_grid=theta_grid,
                    phi_grid=phi_grid,
                )
                # Ylm = Yslm_prec(spin_weight=0, ell=ell, emm=emm, theta=theta_grid, phi=phi_grid)

                test_dist = 3.12345 * Ylm

                test_modes = SHExpand(
                    func=test_dist, method_info=minfo, info=info
                )

                # Test mode coefficient
                np.testing.assert_almost_equal(
                    test_modes.mode(ell, emm),
                    3.12345,
                    13,
                    f"The l{ell} m{emm} mode"
                    "must be 3.12345 upto"
                    "12 digits",
                )

        def test_ylm_multiple_mode_recovery(self):
            """Test multi mode recovery starting with
            a superposition of modes upoto L"""

            info = GLGrid(L=11)
            minfo = method_info(ell_max=11, int_method="GL")

            from spectral.spherical.swsh import Yslm_vec

            theta_grid, phi_grid = info.meshgrid

            Yl0m0 = Yslm_vec(
                spin_weight=0,
                ell=0,
                emm=0,
                theta_grid=theta_grid,
                phi_grid=phi_grid,
            )
            Yl1mm1 = Yslm_vec(
                spin_weight=0,
                ell=1,
                emm=-1,
                theta_grid=theta_grid,
                phi_grid=phi_grid,
            )
            Yl2m2 = Yslm_vec(
                spin_weight=0,
                ell=2,
                emm=2,
                theta_grid=theta_grid,
                phi_grid=phi_grid,
            )
            Yl7m6 = Yslm_vec(
                spin_weight=0,
                ell=7,
                emm=6,
                theta_grid=theta_grid,
                phi_grid=phi_grid,
            )
            Yl9m5 = Yslm_vec(
                spin_weight=0,
                ell=9,
                emm=5,
                theta_grid=theta_grid,
                phi_grid=phi_grid,
            )
            Yl11mm11 = Yslm_vec(
                spin_weight=0,
                ell=11,
                emm=-11,
                theta_grid=theta_grid,
                phi_grid=phi_grid,
            )

            mode_amps = np.random.randn(6)

            mode_amps_dict = {
                "l0m0": mode_amps[0],
                "l1m-1": mode_amps[1],
                "l2m2": mode_amps[2],
                "l7m6": mode_amps[3],
                "l9m5": mode_amps[4],
                "l11m-11": mode_amps[5],
            }

            message(
                "Mode amplitides assumed", mode_amps_dict, message_verbosity=2
            )

            test_data = (
                mode_amps_dict["l0m0"] * Yl0m0
                + mode_amps_dict["l1m-1"] * Yl1mm1
                + mode_amps_dict["l2m2"] * Yl2m2
                + mode_amps_dict["l7m6"] * Yl7m6
                + mode_amps_dict["l9m5"] * Yl9m5
                + mode_amps_dict["l11m-11"] * Yl11mm11
            )

            test_modes = SHExpand(func=test_data, method_info=minfo, info=info)

            # Test the mode coeffs

            rec_amps_dict = {
                "l0m0": test_modes.mode(0, 0),
                "l1m-1": test_modes.mode(1, -1),
                "l2m2": test_modes.mode(2, 2),
                "l7m6": test_modes.mode(7, 6),
                "l9m5": test_modes.mode(9, 5),
                "l11m-11": test_modes.mode(11, -11),
            }

            np.testing.assert_almost_equal(
                rec_amps_dict["l0m0"],
                mode_amps_dict["l0m0"],
                14,
                "The l0m0 mode must be recovered" "to 14 digits",
            )

            np.testing.assert_almost_equal(
                rec_amps_dict["l1m-1"],
                mode_amps_dict["l1m-1"],
                14,
                "The l1m-1 mode must be recovered" "to 14 digits",
            )

            np.testing.assert_almost_equal(
                rec_amps_dict["l2m2"],
                mode_amps_dict["l2m2"],
                14,
                "The l2m2 mode must be recovered" "to 14 digits",
            )

            np.testing.assert_almost_equal(
                rec_amps_dict["l7m6"],
                mode_amps_dict["l7m6"],
                14,
                "The l7m6 mode must be recovered" "to 14 digits",
            )

            np.testing.assert_almost_equal(
                rec_amps_dict["l9m5"],
                mode_amps_dict["l9m5"],
                14,
                "The l9m5 mode must be recovered" "to 14 digits",
            )

            np.testing.assert_almost_equal(
                rec_amps_dict["l11m-11"],
                mode_amps_dict["l11m-11"],
                14,
                "The l11m-11 mode must be recovered" "to 14 digits",
            )

    def test_ylm_contraction(self):
        """Test the data reconstructed from the
        SH expansion against the original data"""

        info = GLGrid(L=11)
        minfo = method_info(ell_max=11, int_method="GL")

        from spectral.spherical.swsh import Yslm_vec

        theta_grid, phi_grid = info.meshgrid

        Yl0m0 = Yslm_vec(
            spin_weight=0,
            ell=0,
            emm=0,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )
        Yl1mm1 = Yslm_vec(
            spin_weight=0,
            ell=1,
            emm=-1,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )
        Yl2m2 = Yslm_vec(
            spin_weight=0,
            ell=2,
            emm=2,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )
        Yl7m6 = Yslm_vec(
            spin_weight=0,
            ell=7,
            emm=6,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )
        Yl9m5 = Yslm_vec(
            spin_weight=0,
            ell=9,
            emm=5,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )
        Yl11mm11 = Yslm_vec(
            spin_weight=0,
            ell=11,
            emm=-11,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )

        mode_amps = np.random.randn(6)

        mode_amps_dict = {
            "l0m0": mode_amps[0],
            "l1m-1": mode_amps[1],
            "l2m2": mode_amps[2],
            "l7m6": mode_amps[3],
            "l9m5": mode_amps[4],
            "l11m-11": mode_amps[5],
        }

        message("Mode amplitides assumed", mode_amps_dict, message_verbosity=2)

        test_data = (
            mode_amps_dict["l0m0"] * Yl0m0
            + mode_amps_dict["l1m-1"] * Yl1mm1
            + mode_amps_dict["l2m2"] * Yl2m2
            + mode_amps_dict["l7m6"] * Yl7m6
            + mode_amps_dict["l9m5"] * Yl9m5
            + mode_amps_dict["l11m-11"] * Yl11mm11
        )

        test_modes = SHExpand(func=test_data, method_info=minfo, info=info)

        test_data_back = test_modes.contract()

        # Test contracted modes
        np.testing.assert_array_almost_equal(
            test_data_back,
            test_data,
            12,
            "The contracted modes must equal"
            " the original input data"
            " upto 12 decimals",
        )

    def test_ylm_reexpand(self):
        """Test the re expansion of the SH modes
        after reconstruction from expanded modes
        against the original mode amplitudes used
        to construct the test data
        (expansion - reconstruction - expansion)"""

        info = GLGrid(L=11)
        minfo = method_info(ell_max=11, int_method="GL")

        from spectral.spherical.swsh import Yslm_vec

        theta_grid, phi_grid = info.meshgrid

        Yl0m0 = Yslm_vec(
            spin_weight=0,
            ell=0,
            emm=0,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )
        Yl1mm1 = Yslm_vec(
            spin_weight=0,
            ell=1,
            emm=-1,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )
        Yl2m2 = Yslm_vec(
            spin_weight=0,
            ell=2,
            emm=2,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )
        Yl7m6 = Yslm_vec(
            spin_weight=0,
            ell=7,
            emm=6,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )
        Yl9m5 = Yslm_vec(
            spin_weight=0,
            ell=9,
            emm=5,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )
        Yl11mm11 = Yslm_vec(
            spin_weight=0,
            ell=11,
            emm=-11,
            theta_grid=theta_grid,
            phi_grid=phi_grid,
        )

        mode_amps = np.random.randn(6)

        mode_amps_dict = {
            "l0m0": mode_amps[0],
            "l1m-1": mode_amps[1],
            "l2m2": mode_amps[2],
            "l7m6": mode_amps[3],
            "l9m5": mode_amps[4],
            "l11m-11": mode_amps[5],
        }

        message("Mode amplitides assumed", mode_amps_dict, message_verbosity=2)

        test_data = (
            mode_amps_dict["l0m0"] * Yl0m0
            + mode_amps_dict["l1m-1"] * Yl1mm1
            + mode_amps_dict["l2m2"] * Yl2m2
            + mode_amps_dict["l7m6"] * Yl7m6
            + mode_amps_dict["l9m5"] * Yl9m5
            + mode_amps_dict["l11m-11"] * Yl11mm11
        )

        test_modes = SHExpand(func=test_data, method_info=minfo, info=info)

        test_data_back = test_modes.contract()

        # Test re expansion
        # Re-expand contracted modes
        test_modes_back = SHExpand(
            func=test_data_back, method_info=minfo, info=info
        )

        test_modes.compare_modes(test_modes_back, prec=12)

    def test_ylm_vs_scipy(self):
        """Test the SH modes upto a given L
        against the scipy package for computing
        SH modes.

        Notes
        -----
        It was observed that the
        scipy package is accurate
        upto 9 digits at :math:`\\ell\\sim 24`

        """

        info = GLGrid(L=24)
        minfo = method_info(ell_max=24, int_method="GL")

        from spectral.spherical.swsh import Yslm_vec

        theta_grid, phi_grid = info.meshgrid

        from scipy.special import sph_harm

        for ell in range(info.L + 1):
            for emm in range(-ell, ell + 1):
                Ylm_this_module = Yslm_vec(
                    spin_weight=0,
                    ell=ell,
                    emm=emm,
                    theta_grid=theta_grid,
                    phi_grid=phi_grid,
                )

                Ylm_scipy = sph_harm(emm, ell, phi_grid, theta_grid)

                # Note: scipy is not accurate. Any disagreement
                # and low precision
                # is due to short comings of scipy's sph_harm

                np.testing.assert_array_almost_equal(
                    Ylm_this_module,
                    Ylm_scipy,
                    9,
                    "The spherical harmonic"
                    f" modes (l{ell}, m{emm}"
                    " from this module"
                    " and scipy must equal"
                    " upto 9 decimals",
                )

    def test_ylm_vs_exact(self):
        """Test the SH modes computed using
        the fast method against the exact
        symbolic computation"""

        # info = GLGrid(L=24)
        # minfo = method_info(ell_max=24, int_method='GL')

        ell_max = 6  # Warning: Very expensive. Tested upto 84

        from spectral.spherical.swsh import Yslm_vec, Yslm_prec

        # theta_grid, phi_grid = info.meshgrid

        theta = np.pi / 21
        phi = np.pi / 6

        for ell in range(ell_max + 1):
            for emm in range(ell, ell + 1):
                message(f"Testing l{ell} m{emm}", message_verbosity=2)

                Ylm_this_module = Yslm_vec(
                    spin_weight=0,
                    ell=ell,
                    emm=emm,
                    theta_grid=theta,
                    phi_grid=phi,
                )

                Ylm_exact = Yslm_prec(
                    spin_weight=0,
                    ell=ell,
                    emm=emm,
                    theta=theta,
                    phi=phi,
                    prec=14,
                )

                np.testing.assert_array_almost_equal(
                    Ylm_this_module,
                    np.complex128(Ylm_exact),
                    12,
                    "The spherical harmonic"
                    "modes from this module"
                    "must agree with exact"
                    "upto 12 digits",
                )

    def test_ylm_vs_exact_grid(self):
        """Test the SH modes computed using
        the fast method against the exact
        symbolic computation"""

        info = GLGrid(L=24)
        # minfo = method_info(ell_max=24, int_method='GL')

        ell_max = 16  # Warning: very expensive. Tested upto 84

        from spectral.spherical.swsh import Yslm_vec, Yslm_prec_grid

        theta_grid, phi_grid = info.meshgrid

        # theta = np.pi/21
        # phi = np.pi/6

        for ell in range(ell_max + 1):
            for emm in range(ell, ell + 1):
                message(f"Testing l{ell} m{emm}", message_verbosity=2)

                Ylm_this_module = Yslm_vec(
                    spin_weight=0,
                    ell=ell,
                    emm=emm,
                    theta_grid=theta_grid,
                    phi_grid=phi_grid,
                )

                Ylm_exact = Yslm_prec_grid(
                    spin_weight=0,
                    ell=ell,
                    emm=emm,
                    theta_grid=theta_grid,
                    phi_grid=phi_grid,
                    prec=14,
                )

                np.testing.assert_array_almost_equal(
                    Ylm_this_module,
                    np.complex128(Ylm_exact),
                    14,
                    "The spherical harmonic"
                    "modes from this module"
                    "must agree with exact"
                    "upto 14 digits",
                )

    def test_ylm_vs_spherical(self):
        """Test the vectorized SH mode
        computation against the spherical
        package modes

        Notes
        -----
        It was observed that the
        spherical package is accurate
        upto 13 digits at :math:`\\l`=100
        """

        def get_index(ell, emm):
            ind = 0
            for ell_ind in range(ell + 1):
                ind += 2 * ell_ind + 1

            return ind + emm - ell_ind - 1

        # info = GLGrid(L=24)
        # minfo = method_info(ell_max=24, int_method='GL')

        ell_max = 24

        from spectral.spherical.swsh import Yslm_vec

        # theta_grid, phi_grid = info.meshgrid

        theta = np.pi / 21
        phi = np.pi / 6
        spin_weight = 0

        import quaternionic, spherical

        R = quaternionic.array.from_spherical_coordinates(theta, phi)
        # ell_max = ell
        wigner = spherical.Wigner(ell_max)
        Y2 = wigner.sYlm(spin_weight, R)

        for ell in range(ell_max + 1):
            for emm in range(ell, ell + 1):
                message(f"Testing l{ell} m{emm}", message_verbosity=2)

                Ylm_this_module = Yslm_vec(
                    spin_weight=0,
                    ell=ell,
                    emm=emm,
                    theta_grid=theta,
                    phi_grid=phi,
                )

                Ylm_spherical = Y2[get_index(ell, emm)]

                np.testing.assert_array_almost_equal(
                    Ylm_this_module,
                    Ylm_spherical,
                    15,
                    "The spherical harmonic"
                    f" l{ell}m{emm} mode"
                    " from the spherical module"
                    " must agree with this"
                    " module upto 14 digits",
                )

    def test_ylm_vs_spherical_grid(self):
        """Test the vectorized SH mode
        computation against the spherical
        package modes

        Notes
        -----
        It was observed that the
        spherical package is accurate
        upto 13 digits at :math:`\\l`=100
        """

        def get_index(ell, emm):
            ind = 0
            for ell_ind in range(ell + 1):
                ind += 2 * ell_ind + 1

            return ind + emm - ell_ind - 1

        info = GLGrid(L=24)
        # minfo = method_info(ell_max=24, int_method='GL')

        ell_max = 24

        from spectral.spherical.swsh import Yslm_vec

        theta_grid, phi_grid = info.meshgrid

        # theta = np.pi/21
        # phi = np.pi/6
        spin_weight = 0

        import quaternionic, spherical

        R = quaternionic.array.from_spherical_coordinates(theta_grid, phi_grid)
        # ell_max = ell
        wigner = spherical.Wigner(ell_max)
        Y2 = wigner.sYlm(spin_weight, R)

        for ell in range(ell_max + 1):
            for emm in range(ell, ell + 1):
                message(f"Testing l{ell} m{emm}", message_verbosity=2)

                Ylm_this_module = Yslm_vec(
                    spin_weight=0,
                    ell=ell,
                    emm=emm,
                    theta_grid=theta_grid,
                    phi_grid=phi_grid,
                )

                Ylm_spherical = Y2.T[get_index(ell, emm)].T

                np.testing.assert_array_almost_equal(
                    Ylm_this_module,
                    Ylm_spherical,
                    14,
                    "The spherical harmonic"
                    f" l{ell}m{emm} mode"
                    " from the spherical module"
                    " must agree with this"
                    " module upto 14 digits",
                )

    @pytest.mark.skip(reason="This tests spherical code. Takes long.")
    def test_spherical_vs_exact_grid(self):
        """Test the vectorized SH mode
        computation against the spherical
        package modes

        Notes
        -----
        It was observed that the
        spherical package is accurate
        upto 13 digits at :math:`\\l`=100
        """

        def get_index(ell, emm):
            ind = 0
            for ell_ind in range(ell + 1):
                ind += 2 * ell_ind + 1

            return ind + emm - ell_ind - 1

        info = GLGrid(L=24)
        # minfo = method_info(ell_max=24, int_method='GL')

        ell_max = 6  # Warning: very expensive. Tested upto 84

        from spectral.spherical.swsh import Yslm_prec_grid

        theta_grid, phi_grid = info.meshgrid

        # theta = np.pi/21
        # phi = np.pi/6
        spin_weight = 0

        import quaternionic, spherical

        R = quaternionic.array.from_spherical_coordinates(theta_grid, phi_grid)
        # ell_max = ell
        wigner = spherical.Wigner(ell_max)
        Y2 = wigner.sYlm(spin_weight, R)

        for ell in range(ell_max + 1):
            for emm in range(ell, ell + 1):
                message(f"Testing l{ell} m{emm}", message_verbosity=2)

                Ylm_exact = Yslm_prec_grid(
                    spin_weight=0,
                    ell=ell,
                    emm=emm,
                    theta_grid=theta_grid,
                    phi_grid=phi_grid,
                    prec=16,
                )

                Ylm_spherical = Y2.T[get_index(ell, emm)].T

                np.testing.assert_array_almost_equal(
                    Ylm_spherical,
                    Ylm_exact,
                    14,
                    "The spherical harmonic"
                    f" l{ell}m{emm} mode"
                    " from the spherical module"
                    " must agree with this"
                    " module upto 14 digits",
                )


# if __name__ == '__main__':

#    unittest.main(argv=['first-arg-is-ignored'], exit=False, verbosity=2)
